#!/usr/bin/env python
# coding=utf-8
# Almut Stribeck almut@stribeck.de
# sacutray.py

# WARNING Workarounds for two shortcomings of matplotlib 2.1.0
#         have been necessary to make the graphics run on newer
# versions of python.
# The problem is, that the new RectangleSelector does NOT return
# the click and release positions as click and release positions,
# but always returns the lower left corner of the "rectangle" as
# pseude-click position and the upper right corner as the pseudo-
# release position. Such a transformation should be left to the
# decision of the user
#
# Moreover, matplotlib 2.1.0 throws exceptions after RectangleSelector
# has finished. Then the TkAgg backend states that "this is no Tk
# application" - at least this warning does not stop the program.

from __future__ import print_function
from sacurve import * # Own module: Curves and primitive operations on them

#import os # For file checks (may be unnecessary here)
import copy # to protect external variables from being altered
import numpy as np # For arrays and their processing
import matplotlib # For checking the backend (useblit)
import matplotlib.pyplot as plt # For plotting of the curves
import sys # To catch keypress events and to know the platform
from pygame import mixer # for error sound
import six # To discriminate between PY2 and PY3

from matplotlib.widgets import AxesWidget, Cursor
import savecsel as vs


class SasPlotClass( object ):
    '''The standard plot for sascur: Inspection of 3 curves and simple
    manipulations (graphics command of the former TOPAS)
    '''

    def __init__( self ):
        '''Define the names of the private status variables of the class
        Ui -- most of these settings could as well have been placed directly
        after the class definition without the prepending self -- see: cmd2.py'''
        self.plotfig = None # The matplotlib figure
        self.ax = None # The (only) graph in the figure
        self.CrossHair = None # Connector ID of the crosshair cursor tool
        self.cid_LineSelector = None # Connector ID of the line selector tool
        self.cidboard = None # Connector-ID of the keyboard event handler
        self.cid_hint = None # Connector-ID of an existing hint text
        self.cid_curve = [] # List of connector-IDs of painted curves
        self.DirtyBlackboard = False # The plot is clean (nothing displayed)
        self.closed = False # The user closed the plot manually. Handle to avoid hanging program
        self.backgr = 0.0 # The background b which is subtracted for display from x
        self.spower = 0 # The power to which x is raised for displaying (y-backgr)*x**spower
        self._selcur = None # A pointer to the selected curve
        self.selected = -1 # No curve selected
        self.markers = None # A list of markers. Invisible curves have the marker ''
        # Set by the keypress event handler and noticed by the line selector callback method
        #     to distinguish among different tasks: What shall be done with the vector given
        #     by the user?
        self.stretch = [1.0, 1.0] # components of the last affine stretching ("a", "x" commands)
        self._linetask = ' '
        self._testing = False # report keystrokes (etc.)
        # Initialize graphical error sound with sound files from the program directory
        mixer.init()
        progdir = sys.path[0]
        self._beep = mixer.Sound(os.path.join(progdir,'sacugraerr.wav')) # activate by: self._beep.play()
        self._beepop = mixer.Sound(os.path.join(progdir,'sacugrapop.wav'))
        self._beeptoolenter = mixer.Sound(os.path.join(progdir,'sacugratoolenter.wav'))
        self._beeptoolexit  = mixer.Sound(os.path.join(progdir,'sacugratoolexit.wav'))
        pass
    # __init__ an instance of SasPlotClass

    def _crosshair(self):
        ''' Defines color and speed of a crosshair cursor.
        '''
        return mycursor
    #END _crosshair() of SasPlotClass

    def _errmsg( self, msg ):
        self._beep.play()
        if sys.platform.startswith('linux'):
            # in bold face
            print('\033[1mERROR:',msg,'\033[0m')
        else:
            print('ERROR:',msg)
    #END _errmsg() of SasPlotClass

    def _warnmsg( self, msg ):
        self._beepop.play()
        print('WARNING:',msg)
        pass
    #END _warnmsg() of SasPlotClass

    #===============BEGIN EVENT HANDLERS========================
    def onkeyboard(self, event):
        '''Handle keyboard events. The keyboard is the primary
        selector of the interactive tasks'''
        sys.stdout.flush()
        if event.key == None:
            #Shift-Lock pressed
            return
        lowerkey = event.key.lower()
        if self._testing:
            print('Keystroke: %s (%s)'%(event.key,lowerkey))
        # Handle matplotlib shortcut for closing the window
        if event.key == 'ctrl+w': #WARNING Manual says ctrl-W would close ALL plots. This is wrong
            self.closed = True # User has closed the window by shortcut. Handle manual exit
            return

        # Do nothing, if a matplotlib toolbar shortcut-key is encountered:
        # WARNING It is assumed that the user has not reconfigured shortcuts in matplotlibrc
        #   zoom, pan, save, fullscreen, fullscreen, grid, log-y, log-x, log-y,, back, back,
        #   forward, forward, home, home, home
        toolbar = plt.get_current_fig_manager().toolbar
        if event.key in ['o','p','s','f','ctrl+f','g','l','L','k','c','left',
                         'v','right','h','r','home']:
            if toolbar.mode != '':
                self._beeptoolenter.play()
            else:
                self._beeptoolexit.play()
            return
        #NOTICE Do not react if a toolbar mode is engaged
        if toolbar.mode != '':
            #WARNING Do not remove the following essential print statement
            msg = 'TOOLBAR BUTTON <%s> ACTIVE. Click to disengage!'%toolbar.mode
            self._errmsg( msg )
            return
        # Ensure that a selected curve still has points
        if self.selected > -1:
            if self.curves[self.selected].n <= 0:
                # erase the curve without points
                self._warnmsg('Selection vanished because OP has become empty')
                self._selredraw() # handles deselect, as well
        # handle events
        if event.key in ['0','1','2','3','4']:
            self.spower = int(event.key)
            # always generate a fresh (full-range) plot
            self.DirtyBlackboard = self._freshplot()
        elif lowerkey in ['ctrl+0','ctrl+1','ctrl+2','ctrl+3']:
            # select a curve for manipulation. ctrl-3 returns to the neutral state..
            newsel = int(lowerkey[-1:]) # New selected curve
            #NOTICE It must be ensured that an empty curve is deselected
            if newsel == 3:
                self.selected = -1 # deselect
                print('Curve selection cancelled')
                self.DirtyBlackboard = self._freshplot()
            else:
                if (self.curves[newsel].n <= 0) or (self.markers[newsel] == ''):
                    self._warnmsg('Empty or invisible OP%i cannot be selected.'%newsel)
                else:
                    if self.selected != newsel:
                        print('OP%i selected. Stretching factors reset'%newsel)
                        self.stretch  = [1.0, 1.0] # reset affine stretching accus
                        self.selected = newsel # store new value
                        self.DirtyBlackboard = self._freshplot()
                    else:
                        print('OP%i remains selected. The accumulated '%newsel)
                        print('stretching factors [x=%7.3f, y=%7.3f] remain available '%(self.stretch[0],self.stretch[1]))
        elif lowerkey == 'a':
            if self.selected < 0:
                self._warnmsg('First select a curve!')
                return
            # scale the selected curve in such way that y0 becomes y1
            self._selcur = self.curves[self.selected]
            print('Sc(a)le selected curve in OP%i'%self.selected)
            print('  Enter a line. The curve is multiplied and "from-y" becomes "to-y"')
            print('  Hint: self.ThePlot.stretch[1] holds the stretching factor')
            self._linetask = lowerkey # Indicate a scaling task
            self.CrossHair.visible = False # Cursor off
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'x':
            if self.selected < 0:
                self._warnmsg('First select a curve!')
                return
            # Stretch or compress the selected curve in x-direction
            self._selcur = self.curves[self.selected]
            print('Stretch/compress horizontally: OP%i'%self.selected)
            print('  Enter a line. The x-interval is multiplied and "from-x" becomes "to-x"')
            print('  Hint: self.ThePlot.stretch[0] holds the stretching factor')
            self._linetask = lowerkey # Indicate a stretch-x task
            self.CrossHair.visible = False # Cursor off
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'd':
            if self.selected < 0:
                self._warnmsg('First select a curve!')
                return
            # Delete points
            self._selcur = self.curves[self.selected]
            print('Delete point(s) from selected curve OP%i'%self.selected)
            print('  enter a line. Its x-range defines the range to delete')
            self._linetask = lowerkey # Indicate a deletion task
            self.CrossHair.visible = False # Cursor off
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'z':
            if self.selected < 0:
                self._warnmsg('First select a curve!')
                return
            # Set y-value to zero
            self._selcur = self.curves[self.selected]
            print('Zero interval in selected curve OP%i'%self.selected)
            print('  enter a line. In its x-range the y-values are set to zero')
            self._linetask = lowerkey # Indicate a zeroing task
            self.CrossHair.visible = False # Cursor off
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'm':
            if self.selected < 0:
                self._warnmsg('First select a curve!')
                return
            self._selcur = self.curves[self.selected]
            print('Move point in selected curve OP%i up or down'%self.selected)
            print('  enter a line')
            self._linetask = lowerkey # Indicate the "move" task
            self.CrossHair.visible = False # Cursor off
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'b':
            print('Enter a latent background subtracted from all curves only for display.')
            print('  enter a line (endpoint y is taken')
            self._linetask = lowerkey # Indicate the background task
            self.CrossHair.visible = False # Cursor off
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'i':
            if self.selected < 0:
                self._warnmsg('First select a curve!')
                return
            self._selcur = self.curves[self.selected]
            print('Insert a new, extra data point in the selected curve OP%i.'%self.selected)
            print('  enter a line (endpoint (x,y) is taken)')
            self._linetask = lowerkey # Indicate the "insert" task
            self.CrossHair.visible = False # ... and activate the line selector
            self.cid_LineSelector.set_active(True)
            #NOTICE the action is performed in self._onselect
        elif lowerkey == 'n':
            # link OP2 to OP1 at the position given
            if (self.curves[1].n > 1) and (self.curves[2].n > 1):
                print('Link curves: Take front part of OP1 and tail of OP2. Result in OP1')
                print('  enter a line. endpoint x  is the border')
                self._linetask = lowerkey # indicate: "link" task
                self.CrossHair.visible = False # ... and activate the line selector
                self.cid_LineSelector.set_active(True)
            else:
                self._warnmsg('Needs curves in OP1 AND in OP2')
        pass
    #END onkeyboard() of SasPlotClass

    def _onselect(self, linefrom, lineto, roi):
        '''For the line selector in ScatKnife. This method serves all the
        tasks which are triggered by keyboard events caught by the onkeyboard
        procedure. The defining keystroke is in self.linetask. This is already
        lowercase!
        '''
        aborted = linefrom is None
        if aborted:
            self._beep.play()
            self.cid_LineSelector.set_active(False)
            self.CrossHair.visible = True
            return
        x1, y1 = linefrom.xdata, linefrom.ydata # in data coordinates
        x2, y2 = lineto.xdata, lineto.ydata
        y1 = self._narty( x1, y1 ) # undo display transform
        y2 = self._narty( x2, y2 )
        # Switch off the line selector
        self.cid_LineSelector.set_active(False)
        self.CrossHair.visible = True
        if self._linetask == 'a':
            # scale selected curve by factor f in y-direction
            if (y1 != 0.0) and (y2 != 0.0):
                f = y2/y1
            else:
                f = 1.0
            self._selcur.y = self._selcur.y * f # vector operation
            self.stretch[1] = self.stretch[1]*f # cumulate stretching factor
            # Redraw the transformed selected curve
            self._selredraw()
        elif self._linetask == 'x':
            # stretch or compress the curve in x-direction
            if (x1 != 0.0) and (x2 != 0.0):
                f = x2/x1
            else:
                f = 1.0
            self._selcur.x = self._selcur.x * f # vector operation
            self.stretch[0] = self.stretch[0]*f # cumulate stretching factor
            # Redraw the transformed selected curve
            self._selredraw()
        elif self._linetask == 'd':
            # *** delete points from selected curve
            # Arrange the points in ascending order
            minx = min( x1, x2 )
            maxx = max( x1, x2 )
            # Compute the indices of the points which are closest
            # to the support ends of the line entered by the user
            n1 = self._selcur._catchpos( minx )
            n2 = self._selcur._catchpos( maxx )
            if n2 > n1:
                # May be the range must be shortened
                if self._selcur.x[n1] < minx:
                    # Left caught point was outside the line
                    n1 = n1+1
                if maxx < self._selcur.x[n2]:
                    # Right caught point was outside the line
                    n2 = n2-1
                #NOTICE The result should still fulfil n1 <= n2
            # print('Delrange: (%i,%i)'%(n1,n2))
            self._selcur.delirange(n1, n2)
            # Redraw the selected curve
            self._selredraw()
            pass # task was: delete points
        elif self._linetask == 'n':
            # Link curve in OP2 to curve in OP1
            # Here n1 becomes the cutoff-index in OP1, and n2
            # the cutoff index in OP2
            n1 = self.curves[1]._catchpos( x2 )
            if self.curves[1].x[n1] > x2:
                n1 = n1 - 1
            n2 = self.curves[2]._catchpos( x2 )
            if self.curves[2].x[n2] <= x2:
                n2 = n2 + 1
            #print('xcut=%8.3e'%x2)
            #print('OP1: n1=%i, xhi=%8.3e'%(n1,self.curves[1].x[n1]))
            #print('OP2: n2=%i, xlo=%8.3e'%(n2,self.curves[1].x[n2]))
            coplen =  self.curves[2].n - n2
            newpoints = n1+1+coplen
            if newpoints <= self.curves[1]._maxpoints:
                # mnodify OP1
                self.curves[1].x[n1+1:n1+1+coplen] = self.curves[2].x[n2:n2+coplen]
                self.curves[1].y[n1+1:n1+1+coplen] = self.curves[2].y[n2:n2+coplen]
                self.curves[1].n = newpoints
                self._opredraw(1) # redraw the modified curve
                #print('source: ', self.curves[2].x[n2:n2+coplen] )
                #print('copied: ', self.curves[1].x[n1+1:n1+1+coplen] )
            else:
                self._beep.play()
                print('+++ IMPOSSIBLE: Linked curve would become too long')
        elif self._linetask == 'm':
            # *** Move points up or down in selected curve
            # grab the point
            grab = self._selcur._catchpoint( x1, y1 )
            # change y value
            self._selcur.y[grab] = y2
            # redraw
            self._selredraw()
            pass
        elif self._linetask == 'z':
            # *** Zero all points in the x-range of the vector
            # Arrange the points in ascending order
            minx = min( x1, x2 )
            maxx = max( x1, x2 )
            # Compute the indices of the points which are closest
            # to the support ends of the line entered by the user
            n1 = self._selcur._catchpos( minx )
            n2 = self._selcur._catchpos( maxx )
            if n2 > n1:
                # May be the range must be shortened
                if self._selcur.x[n1] < minx:
                    # Left caught point was outside the line
                    n1 = n1+1
                if maxx < self._selcur.x[n2]:
                    # Right caught point was outside the line
                    n2 = n2-1
            length = n2 - n1 + 1
            # set the y-values in the range to zero
            mask = np.zeros(length)
            self._selcur.y[n1:n2+1] = mask
            # and redraw
            self._selredraw()
            pass
        elif self._linetask == 'b':
            # *** Define a new latent background
            # The release point y2 becomes the new background
            self.backgr = y2
            print('Displayed background is now: %10.3e'%self.backgr)
            # redraw
            self.DirtyBlackboard = self._freshplot()
            #TODO May be one should ask, if only a single curve is shown and,
            #     in this case, one could assign the last backgr to the curve
            pass
        elif self._linetask == 'i':
            if self._selcur._maxpoints == self._selcur.n:
                self._beep.play()
                print('+++ Insertion impossible: Curve has already %i points.'%self._selcur._maxpoints)
                return
            # find the insertion position
            points = self._selcur.n
            nn = self._selcur._catchpos( x2 )
            if self._selcur.x[nn] == x2:
                self._beep.play()
                print('+++ Requested insertion exactly on existing point is impossible')
                return
            if self._selcur.x[nn] < x2:
                nn = nn+1 # nn may become n --> simple append
                pass
            # Point to be inserted must be put at index nn
            if nn < points:
                # make space for the point
                upx = self._selcur.x[nn:points]
                self._selcur.x[nn+1:points+1] = upx
                upy = self._selcur.y[nn:points]
                self._selcur.y[nn+1:points+1] = upy
            # insert the new point
            self._selcur.x[nn] = x2
            self._selcur.y[nn] = y2
            # bookkeeping
            self._selcur.n = points + 1
            # and redraw
            self._selredraw()
        pass
    #End _onselect() of SasPlotClass
    #===============END   EVENT HANDLERS========================

    #===============BEGIN internal methods========================
    # Some of these methods are also called by event handlers

    def _narty( self, xp, yp ):
        ''' Back-transformation of a single y-value according to the actual values of
        self.spower and self.backgr. The forward transform "_trany" is more effectively
        coded in 2 operations on complete vectors'''
        res = 0.0
        if xp > 0.0:
            res = yp / xp**self.spower
            res = res + self.backgr
        return res
    #END _narty() of SasPlotClass

    def _axnice( self, lo, hi ):
        '''extends the given interval (lo,hi) somewhat
        '''
        interval = hi - lo
        if lo > 0.2*interval:
            # extend downwards beyond zero
            lo = min( lo, 0.0 )
            interval = hi - lo
            pass
        eps = 0.02*interval
        lon = lo - eps
        hin = hi + eps
        return lon, hin
    #End _axnice() of SasPlotClass

    def _opredraw( self, idx ):
        '''Redraw the curve with OP-index idx'''
        if self.cid_curve[idx] != None:
            # remove old curve from plot
            self.cid_curve[idx].set_visible(False)
            self.plotfig.canvas.draw()
            self.cid_curve[idx] = None

        if self.curves[idx].n > 0:
            # There are (now) points in the curve: draw it!
            actcur = self.curves[idx] # point at the curve
            points = actcur.n
            if self.markers[idx] != '':
                # There are points and a marker associated to the curve
                x = actcur.x[:points] # copy of the valid data
                y = actcur.y[:points]
                # Do the "_trany"
                y = y - self.backgr
                y = y * ( x**self.spower )
                col = 'k' # default: black markers
                if idx == self.selected:
                    col = 'r' # red markers
                self.cid_curve[idx] = plt.scatter(x, y, marker=self.markers[idx], color=col)
                self.plotfig.canvas.draw() # Show everything
                pass # There is a marker character assigned to this curve
            pass # There are points in this curve
        pass
    #END _opredraw() of SasPlotClass

    def _selredraw( self ):
        '''Redraw the selected curve'''
        if self.curves[self.selected].n == 0:
            if self.cid_curve[self.selected] == None:
                # Nothing to do: curve is already invisible
                return
            # Only erase the curve, if it has no points any more
            self.cid_curve[self.selected].set_visible(False)
            self.plotfig.canvas.draw()
            self.cid_curve[self.selected] = None
            self.selected = -1
        else:
            self._opredraw( self.selected )
    #End _selredraw() of SasPlotClass

    def _freshplot( self ):
        '''Clean blackboard, if it is dirty and plot
        '''
        if self.closed:
            return False # The plt is dead. Exit to prevent a hanging program.
        followup = self.DirtyBlackboard
        if self.DirtyBlackboard:
            # NOTICE matplotlib 1.3 is not designed for proper cleaning and
            #        redrawing in the same plot. If one tries, the attached
            # tools (crosshair, LineSelector) get lost. Therefore: Brute force
            plt.close()
            self.DirtyBlackboard = False

        # NOTICE The following 3 lines had been in the wrapper (ScatKnife), but
        #        when we kill the plot before presenting a follow-up they must be
        #        here.
        self.plotfig = plt.figure() # Generate a matplotlib figure
        # Now subdivide the plotting area into a one-by-one matrix and address its first
        # element as the active plotting region (well, this is the trivial case):
        self.ax = self.plotfig.add_subplot(1,1,1) # argu also 111. raster 1,1 and 1st subplot
        # connect to event handler(s)
        # WARNING NEVER connect a keyboard handler to something below the complete figure.
        #         If you do, then keys may be swallowed -- in particular the "ctrl-w" of
        # which indicates that the figure has been closed, which needs special treatment.
        self.plotfig.canvas.mpl_connect('key_press_event', self.onkeyboard)

        ''' This is what makes problems unless special cases (in particular close_event )
        are not considered:
        self.cidboard = AxesWidget(self.ax).connect_event('key_press_event', self.onkeyboard)
        Then all events can be disconnected by a single call:
            AxesWidget(self.ax).disconnect_events()
        When we connect on the figure level as is shown below, disconnecting is
        performed cid by cid:
            cidboard = self.plotfig.canvas.mpl_connect('key_press_event', onkeyboard)
            self.plotfig.canvas.mpl_disconnect(cidboard) # would disconnect'''

        SomethingPainted = False
        # Initiate tools for interactive work
        blit = matplotlib.get_backend() == 'TkAgg' #TkAgg knows how to blit
        if blit:
            col = 'green'
        else:
            col = 'red'    # slow cursor
        self.CrossHair = Cursor(self.ax, useblit=blit, color=col, linewidth=1)
        self.cid_LineSelector = vs.VSelector(self.ax,self._onselect, 'line', blit)
        self.cid_LineSelector.set_active(False) # Line selector inactive

        roundedpower = int(self.spower + 0.5)
        self.hint = '[$(y-b)s^%s$] $s^{key}$: 0-4. Sel: ctrl+0...+3, b, sc(a)le, x, li(n)k, (d)el, (i)ns, (m)ov, (z)ero, '%roundedpower
        self.cid_hint = plt.figtext(0.01, 0.99, self.hint, style='normal',
                                    verticalalignment='top',fontsize=14*self.ffc, bbox={'facecolor':'green', 'alpha':0.3, 'pad':5})

        submitted = len(self.curves) # Number of curves which have been submitted for plotting
        self.cid_curve = [] # takes the connector ids of a presented curve
        # for computing of the viewport
        for i in range(submitted): # Preset: No connectors
            self.cid_curve.append( None )
        firstcurve=True
        # Preset label strings in LaTeX notation.
        # Will be overwritten by respective curve attributes,
        # and the last set of labels will persist.
        xls, xus = '$x$', 'a.u.'
        yls, yus = '$y$', 'a.u.'
        for i in range(submitted):
            actcur = self.curves[i]
            points = actcur.n
            if (points > 0) and (self.markers[i] != ''):
                # curve can be presented
                # Update label strings, if attribute keys are present
                xls = actcur.attribute('axis1',None,xls)
                xus = actcur.attribute('units1',None,xus)
                yls = actcur.attribute('axis2',None,yls)
                yus = actcur.attribute('units2',None,yus)
                # process curve data
                x = actcur.x[:points] # only valid data
                y = actcur.y[:points]
                y = y - self.backgr
                y = y * ( x**self.spower )
                col = 'k' # Default color: black
                if  i == self.selected:
                    col = 'r' # selected curve has markers in red
                    # WARNING Only few markers can be painted in red in the old matplotlib
                self.cid_curve[i] = plt.scatter(x, y, marker=self.markers[i], c=col)
                if firstcurve:
                    xmax = max(x)
                    xmin = min(x)
                    ymax = max(y)
                    ymin = min(y)
                    firstcurve = False
                else:
                    if max(x) > xmax:
                        xmax = max(x)
                    if min(x) < xmin:
                        xmin = min(x)
                    if max(y) > ymax:
                        ymax = max(y)
                    if min(y) < ymin:
                        ymin = min(y)
                SomethingPainted = True
                pass
            pass
        # And activate the graphics
        # NOTICE we must set the status before we enter plt.show, because the
        #        interactive session is inside plt.show. When it leaves we are done
        self.DirtyBlackboard = SomethingPainted
        if SomethingPainted:
            # bound may be changed by the user, lim may be not.
            xmin, xmax = self._axnice( xmin, xmax )
            ymin, ymax = self._axnice( ymin, ymax )
            # Generate LaTeX string for y-label
            if (self.backgr == 0.0) and (self.spower == 0):
                lyl = '%s [%s]'%(yls,yus)
            elif self.backgr == 0:
                lyl = '%s %s$^{%i}$ [%s]'%(yls,xls,self.spower,yus)
            else:
                lyl = '(%s - %8.2e) %s$^{%i}$ [%s]'%(yls,self.backgr,xls,self.spower,yus)
            plt.ylabel(lyl,fontsize=18*self.ffc)
            lxl = '%s [%s]'%(xls,xus)
            plt.xlabel(lxl,fontsize=18*self.ffc)
            plt.axis([xmin, xmax, ymin, ymax])
            # Show the origin axes, if they are inside the figure
            self.ax.axhline(y=0, color='k')
            self.ax.axvline(x=0, color='k')
            # switch to scientific format between 10^{-3} and 10^{+3}
            self.ax.ticklabel_format(style='sci', axis='x', scilimits=(-3,3))
            self.ax.ticklabel_format(style='sci', axis='y', scilimits=(-3,3))
            plt.show()
        return SomethingPainted
    #END _freshplot() of SasPlotClass

    #===============END   internal methods========================

    #===============BEGIN PUBLIC METHODS========================
    def ScatKnife(self, curves, markers, fontfactor, backgr=None, spower=None):
        '''The "swiss knife" of interactive graphics for a set of
        scattering curves. The curves are shown simultaneously in :-)
        scatter plots
            curves:  a list of CurveClass curves which may become modified
            markers: a corresponding list of markers to be used for each curve.
                 A marker '' tells that this curve shall not be displayed.
                 Only few markers can be painted in color, if matplotlib 1.3 is used
            fontfactor: A factor that scales the sizes of the label fonts used
                 in the plot. Unfortunately matplotlib 1.x has smaller fonts than
                 matplotlib 2.x -- We take care for some balancing
                 Hint: In this program the user can set the fontfactor at runtime:
                 "set fontfactor 0.7"
            backgr:  Just for display transform (x,y) --> (x,y') with y'=y-backgr
            spower:  instead of (x,y') show (x, x^spower * y') (Porod, Laplace ...)
        '''
        # --- Arguments: if necessary set defaults
        if backgr is None:
            backgr = 0.0
        if spower is None:
            spower = 0.0

        # --- init the plot
        self.DirtyBlackboard  = False #NOTICE Wiping not yet implemented

        # write state into instance
        self.backgr = backgr
        self.spower = spower
        self.ffc = fontfactor
        self.curves = curves
        self.markers = markers
        # When a curve is stretched interactively by (a) or (x),
        # the cumulative stretchfactors of the last selected curve
        # are presented in self.stretch (From sascur: self.ThePlot.stretch)
        self.stretch = [1.0, 1.0]
        #roundedpower = int(self.spower + 0.5)
        #self.hint = '[$(y-b)s^%s$] $s^{key}$: 0-4. Sel: ctrl+0...+3, b, sc(a)le, li(n)k, (d)el, (i)ns, (m)ov, (z)ero, '%roundedpower
        # Call the plotter -- a fresh one for every new view.
        self.DirtyBlackboard = self._freshplot()
        self.closed = False
        # The garbage collector will clean up when the graph is closed
    #END ScatKnife() of SasPlotClass

    #===============END   class SasPlotClass ========================

class IaSmoothPlot( object ):
    '''An interactive window to inspect the spline smoothing of curves'''

    def __init__( self ):
        '''Define the names of the private status variables of the class'''
        self.plotfig = None # The matplotlib figure
        self.ax = None # The (only) graph in the figure
        self.cidboard = None # Connector-ID of the keyboard event handler
        self.cid_hint = None # Connector-ID of an existing hint text
        self.cid_curve = [] # List of connector-IDs of painted curves
        self.closed = False # The user closed the plot manually. Handle to avoid hanging program
        self._testing = False # report keystrokes (etc.)
        self._pressedkey = ''
        # Initialize graphical error sound with sound files from the program directory
        mixer.init()
        progdir = sys.path[0]
        self._beep = mixer.Sound(os.path.join(progdir,'sacugraerr.wav')) # activate by: self._beep.play()
        self._beepop = mixer.Sound(os.path.join(progdir,'sacugrapop.wav'))
        self._beeptoolenter = mixer.Sound(os.path.join(progdir,'sacugratoolenter.wav'))
        self._beeptoolexit  = mixer.Sound(os.path.join(progdir,'sacugratoolexit.wav'))
        pass
    # __init__ an instance of IaSmoothPlot

    def onkeyboard(self, event):
        '''Handle keyboard events of IaSmoothPlot.'''
        sys.stdout.flush()
        if event.key == None:
            #Shift-Lock pressed
            return
        lowerkey = event.key.lower()
        if self._testing:
            print('Keystroke: %s (%s)'%(event.key,lowerkey))
        # Handle matplotlib shortcut for closing the window
        if event.key == 'ctrl+w': #WARNING Manual says ctrl-W would close ALL plots. This is wrong
            self.closed = True # User has closed the window by shortcut. Handle manual exit
            return

        # Do nothing, if a matplotlib toolbar shortcut-key is encountered:
        # WARNING It is assumed that the user has not reconfigured shortcuts in matplotlibrc
        #   zoom, pan, save, fullscreen, fullscreen, grid, log-y, log-x, log-y,, back, back,
        #   forward, forward, home, home, home
        toolbar = plt.get_current_fig_manager().toolbar
        if event.key in ['o','p','s','f','ctrl+f','g','l','L','k','c','left',
                         'v','right','h','r','home']:
            if toolbar.mode != '':
                self._beeptoolenter.play()
            else:
                self._beeptoolexit.play()
            return
        #NOTICE Here the simple user keystroke does not interfere with the toolbar modes. So
        #       the keystrokes shall always be served.
        if lowerkey in ['y','j']:
            self._pressedkey = 'y'
        elif lowerkey == 'n':
            self._pressedkey = 'n'
        else:
            self._beep.play() # "I do not understand this key"
            self._pressedkey = '' #irrelevant
        if self._pressedkey != '':
            plt.close() # A valid answer is in _pressedkey. So close the plot!
        pass
    #END onkeyboard event handler

    def _showsmoothing( self, curve, ffc, sp, equi ):
        '''Shows the raw data and the smoothing of the curve and asks for acceptance.
        ffc is the fontfactor -- a settable variable of the cmd2 environment.'''

        self.plotfig = plt.figure() # Generate a matplotlib figure
        self.ax = self.plotfig.add_subplot(1,1,1) # argu also 111. raster 1,1 and 1st subplot
        # connect to event handler(s)
        self.cidboard = AxesWidget(self.ax).connect_event('key_press_event', self.onkeyboard)
        self.hint = 'Smoothing o.k (y/n)?'
        self.cid_hint = plt.figtext(0.01, 0.98, self.hint, style='normal',
                                    verticalalignment='top',fontsize=14*ffc, bbox={'facecolor':'green', 'alpha':0.3, 'pad':5})
        n = curve.n
        plt.plot(curve.x[:n],curve.y[:n], 'b.') #rough: blue dots
        smo = curve._spline( sp, equi )
        n = smo.n
        plt.plot(smo.x[:n], smo.y[:n], 'r', linewidth=2) #smooth: red, thicker line
        plt.show()
        # The plot is closed in onkeyboard when a valid answer has arrived in self._pressedkey
        return self._pressedkey
    #END _showsmoothing() of IaSmoothPlot

#===============END   class IaSmoothPlot =======================================
