#!/usr/bin/env python
# -*- coding: utf-8 -*-
# coding=utf-8
# Almut Stribeck almut@stribeck.de
'''
sapatt.py -- Define n-dimensional data structures for the small-angle
             scattering, windows which present them and interactive
operations on them. In this (first) version only one 2D scattering pattern is
defined. More complex patterns may be children of this basic "Patt" class.

TESTED
    Python 2.7.6 (default, Oct 26 2016, 20:32:47) [GCC 4.8.4]
        in conjunction with matplotlib 1.3.1 on Linux (Kubuntu),
    Python 3.6.3 (compiled from sources on Linux (CentOS)) [GCC 4.7.7]
        in conjunction with matplotlib 2.1.0

    WARNING For MS Windows the Python versions > 3.5.1 are broken!
            No interactive windows will work, because of a
            fundamental bug introduced into cpython. As soon as
            an INTERACTIVE window is opened, it shows the message
            "(Not Responding)" - and the program hangs.
                Cf.: http://bugs.python.org/issue31546
                PyOS_InputHook is not called when waiting for input()
USAGE
    import sapatt as sp # "scattering pattern"
    a = sp.Patt()
    a.load('mypatt.sjs') # Loads the pattern from a file

EXTERNAL REQUIREMENTS
    For (p)lotting in (i)nter(a)ctive (windows) pattclass relies on my own
    module  piawindows -- and piawindows uses my module vecsel for all the
    mouse operations. Every mouse operation is, on an abstract level, the
    selection of a vector in the open window.

If interactive windows are addressed via sp.showi. sapiawin must be imported:
    import sapiawin as pw
    self.pia_win_d = {} # Generate an empty dictionary for interactive windows
    pia = pw.piawins( self.pia_win_d, verbose=True ) # Make the window handler
    a = sp.Patt()
    a.load('mypattern')
    a.showi(pia, wintit='mypattern', ulev=0.3, clut='gnuplot') # Show pattern

    With an open window there are several draw-related attributes, e.g.
    a.drawwin # The window number of the window in which the pattern is presented

    For the animated presentation of an ongoing background evaluation one
    should call the pia.popup and pia._freshup directly with copies of the
    mapps. Only use NOTICE draw = a.copy() NOTICE  to generate a copy!!!
'''
from __future__ import print_function
import sys, os, six # For compatibility checks
from matplotlib.path import Path
import matplotlib.pyplot as plt

# *** Saving and loading in various file formats
import xdrlib, struct    # For SUN X-platform binary files (PV-WAVE, IDL)
if six.PY2:
    import cPickle as pickle # The python standard: fast, dangerous
else:
    import pickle # Now it is simply called pickle

# * json_tricks can store and restore much more complex objects
# * than the json standard and does it in native python types.
# * The profit is complete identity after a save/load, the price
# * is compatibility and readability by other languages.
#         for Python 2.7: pip install 'json-tricks<3.0' else:
#         from Python 3:  pip install json-tricks  NOTICE
import json_tricks as jst # relatively fast and secure

# * No, we cannot use json_tricks with keyword primitives=True
# * to do the job of json, because this keyword is only available
# * for versions > 3.7. Moreover, the original is faster.
# * json cannot store numpy arrays in their native form. So the
# * arrays must be converted to lists before saving and after
# * loading they must be converted back to ndarrays.
import json

# *** For the core business of pattern and image processing
import numpy as np
from scipy import ndimage

# WARNING Original and copy will not always be completely decoupled
import copy # for copies of instances

class Patt( object ):
    '''A 2D scattering pattern or detector image.
All the persistent data of the pattern must be placed in its "data dictionary"
Patt._dd, otherwise they will not be saved to file or loaded from file:
    The data dictionary has data stored under 3 keys:
        Patt._dd['env'] is the place to store all the attributes
                        (environmental data) related to the pattern
        Patt._dd['map'] contains the central data of the pattern, i.e. the
                        values of the studied pattern at each of its pixels.
                        It is represented by a 2D array of np.float32
                        dtype.
        patt._dd['proc_hist'] a list of strings which may be used to document
                              the processing history of the pattern.
                              Do not use "#" or "//", because json would treat
                              the rest as comment and remove it before save.
    Important parameters of the pattern are accessible "as if they were
    variables" by setter/getter methods. The setters check the validity of
    the user input.

    The CONSTRUCTOR of the pattern has only two parameters: the map itself
    and a "function descriptor" (fund), which adds to the minimum attribute
    set which is necessary for a Patt to work.

    Deliberate attributes that describe the pattern can be defined dynamically
    using the method Patt.puta(). The keys of these attributes must be
    strings. This is not a restriction put by Python, but guarantees that
    patterns can be saved to files in almost every basic file format scheme.

    A FILE FORMAT READER should be split into two parts:
    (a) A file format interpreter. This method should return the map
        and a dictionary of attributes read from the file (attribute keys
        as strings)
    (b) A Patt feeder. This method must convert the map to proper type
        and must translate the read dictionary entries into the appropriate
        dictionary entries which are understood by Patt. Of course, it must
        take care that at least the essential attributes are set.

    Usage: (a) Generate a Patt from experimental data, e.g. PILATUS tiff format
               import samonotiff as mt # My own tiff-reader for detector images
               import sapatt as sp

               ## Read the tiff-file into a numpy matrix m
               ## and a dictionary t which contains the tag list
               m, t = mt.tiffread('rd15dd_sa_00001_00019.tif')

               ## Generate an instance of a scattering pattern named "a",
               #  fill it from the numpy-array "m" and supply a simple
               #  function descriptor made from the value stored under
               #  the key "x_micron" divided by 1000 and the unit "mm".
               #  This indicates quadratic pixels in mm, as is usual
               #  for detector images. (the numerical value is stored
               ## in a_dd['env']['_vaxras'])
               a = sp.Patt(m, (t["x_micron"]/1000, 'mm'))
               ## Translate the tif-dictionary entry 'document_name'
               ## into a Patt attribute 'title' and enter the value.
               a.puta( 'title', t["document_name"] ) # put a title
               ## Is there a "title" present? (returns "True" or "False")
               a.hasa('title') # Print the answer to the screen
               ## Print all the attributes and their values stored in a
               a.hasa()
               ts = a.geta('title')  # Put the title into a variable ts
               a.mapp # The pattern matrix
                   (WARNING "map" is a Python reserved word)
               a.height # The image height
               a.width # The image width
               # vaxvars, vaxunits, vaxras are properties of the pattern
               # which can be addressed as if they were variables. They
               # are vectors with (at least) 3 components related to
               # the description of axes (at least x-, y-, function-).
               # The components of vaxvars and vaxunits are TeX strings
               # in which curly braces are replaced by double curly braces
               # for correct interpretation by matplotlib.
               a.vaxvars   # names of the variables (access via property)
               a.vaxunits  # .. their physical properties
               a.vaxras    # the pixel lengths and a flag
           (b) save a Patt to a file the name of which is
                   made from an attribute "fn" and the suffix '_final':
               a.save( a.get['fn']+'_final' )
           (c) load a Patt from a file
               import sapatt as sp # always necessary
               b = sp.Patt() # generate a primitive Patt
               # fill it from a pickle file (extension: .spp)
               b.load( 'rd15dd_sa_00001_00019_final' )'''
    #END docstring of Patt


    def __init__( self, mat=None, fund=None ):
        """Typical use:
    import sapatt as sp
    a = sp.Patt() # create dummy instance (for fill from file)
    a = sp.Patt( mymap ) # Without setting a function descriptor
    a = sp.Patt( mymap, (0.172, "mm") ) # Technical info PILATUS
    # The last call generates the function descriptor for an
    # intensity in the detector plane I(x,y) with a pixel size
    # of 0.172 mm.

    # For loading mymap from a PILATUS-Tiff you may start with:
    import Image # If you want to use the standard python library
    mymap = np.float32( Image.open( "myexp_00001_00019.tif") )
    # ... or use my own samonotiff.py that does not flip high
    #     intensities into negative values:
    import samonotiff as mt
    mymap, t = mt.tiffread('rd15dd_sa_00001_00019.tif')

    Full description of the constructor call:
    a = sp.Patt( mymap [, fund ] )

        mymap : The image of the scattering pattern supplied as a numpy
                matrix of dtype float32, e.g.
                m = np.float32([ [1, 2], [3, 4] ])

        fund :  Description of the image map in terms of a two-dimensional
                function. If fund is not provided, the required data must
                be placed by a method that supplies the primitive
                environmental data of the experiment later.

                Defining the short form of fund:
                f = (0.172, "mm") # The pixel size and unit

                The function descriptor part of _makepatt defines 3
                attributes of the instance by setter methods which check
                the validity of the right side of the equation.
                a.vaxvars = ["$r_{{12}}$","$r_3$","$z$"]
                a.vaxunits = ["nm","nm","a.u."]
                a.vaxras  = [0.01,0.01,0]
        """
        # Generate the "internal data dictionary"
        # NOTICE may be defined as "Ordered" dictionary
        self._dd = {}
        # fill it by _makepatt()
        self._makepatt( mat, fund )
        # notepad variables
        self.lastcenter = None # Last center of a circle input
        self.lastline = None # Result of last "line" input (x1,y1,x2,y2)
    #END constructor __init__() of Patt

    def perror(self, errmsg):
        # Error message printing using the same syntax as that of module cmd2
        err = "ERROR: {}\n".format(errmsg)
        sys.stderr.write( err )
    #END perror() of Patt

    def isstring( self, item ):
        # work around the missing unicode type in Python3
        if six.PY2:
            return isinstance( item, (str, unicode) )
        else:
            return isinstance( item, str )
    #END isstring() of Patt

    def copy(self):
        """Return a copy of the actual pattern - a painful exercise because
        of errors in the copy module."""

        # res = copy.deepcopy( self ) # WARNING This does not work
        # NOTICE deepcopy must be replaced by a sequence of copies to avoid the
        #        error "TransformNode instances can not be copied". Moreover,
        # if it does not throw an exception on some further manipulation, a
        # change of the mapp of the copy will cause strange changes in the
        # original (even complete erasure of the mapp). The hint to use frozen()
        # does not work.

        # NOTICE Even the following code does not decouple the mapps of self and
        #        of res. Both pointers still point at the same array (Python 2.7).
        #        This is a fundamental programming error in the copy module.
        #res = copy.copy( self ) # WARNING This does not work
        #res._dd['map'] = copy.copy(self._dd['map'])

        # The last resort: Copy by implanting the mapp into a new instance
        implant = copy.copy(self.mapp)
        res = Patt(implant)

        # ... and copy the pattern attributes and its history list
        res._dd['env'] = copy.copy(self._dd['env'])
        res._dd['proc_hist'] = copy.copy(self._dd['proc_hist'])
        # NOTICE if the definition of Patt is changed, this method must be
        #        adapted, as well.
        return res
    #END copy() of Patt

    def _makepatt( self, mat=None, fund=None ):
        '''Generate a primitive pattern by storing its 2D matrix
        "mat" (which represents the function) together with a
        function descriptor in the self._dd dictionary
        '''
        # Generate an attribute dictionary for the pattern
        self._dd['env'] = {}
        # Generate an empty processing history list
        self._dd['proc_hist'] = []

        # Generate the self._dd.['map'] with the intrinsic
        # attributes "width" and "height"
        if mat is None:
            # An empty, tiny 2D matrix as a default
            mat = np.float32([ [0,0,0], [0,0,0], [0,0,0] ])

        # Direct call of the user-callable setter which only works
        # if a matrix is supplied which can be used as a map.
        self._setthemap( mat )

        # Construct a function descriptor
        # Default: It's a scattering pattern fresh from a PILATUS
        # detector with a pixel size of 172 x 172 micrometers.
        # The variable names are written in LaTeX convention starting
        # from normal "text mode" (as units generally are written)
        # NOTICE matplotlib displays single curly braces, but double
        #        braces are handed to TeX
        # NOTICE upright characters inside a "dollar clause: \mbox{{nm}}
        #        bold vectors: \mathbf{{}} TODO: special treatment for backslashs?
        vaxvars = ["$x$","$y$","$I$"]  # List of variable names in TeX convention
        vaxunits = ["mm","mm","a.u."]
        vaxras  = [0.172,0.172,1]
        if fund is None:
            funset = (vaxvars, vaxunits, vaxras)
        else:
            # Called with 2 parameters
            if not isinstance( fund, tuple ):
                self.perror("Function descriptor no tuple: Put in ()")
                return True
            if isinstance( fund[0], (int, long, float) ):
                # short version of the fund: check values
                # do we have two values
                if len(fund) != 2:
                    self.perror('Short "fund" tuple must have 2 members (pixsize, pixunit).')
                    return True
                if fund[0] <= 0:
                    self.perror('Pixel size must be positive')
                    return True
                if not self.isstring( fund[1] ):
                    self.perror('Pixel unit must be a string. "Quote"!')
                    return True
                myunit = fund[1].strip()
                if len(myunit) < 1:
                    self.perror('Pixel unit string must not be empty')
                    return True
                # Modify the default lists: identical pixel units, quadratic pixels
                vaxunits[0], vaxunits[1] = myunit,  myunit
                vaxras[0],  vaxras[1]  = fund[0], fund[0]
            else:
                # Check long version of fund
                if not all(isinstance(item, list) for item in fund):
                    self.perror("All items in long description must be lists: []")
                    return True
                # Rest of the checking is handed out to the list setters
                vaxvars = fund[0]
                vaxunits = fund[1]
                vaxras  = fund[2]
            #End discrimination: fund provided

            # Call the attribute setters
            self._setvaxvars( vaxvars )
            self._setvaxunits( vaxunits )
            self._setvaxras( vaxras )
            # If the last item of vaxras is positive, then all the function
            # values in the map must be positive. Then remove negative
            # values which have been passed in the matrix!
            if vaxras[-1] > 0:
                # ... then "0" means: undefined.
                self.mapp[ self.mapp < 0 ] = 0.0 # getter/setter for clarity
        #END _makepatt() of Patt

# START ++++++++++Basic attribute management+++++++++++++++++++++++++++++++++++++++++

    def puta( self, key, value ):
        '''Put or replace an attribute in the attribute dictionary
    img._dd['env']  Call: img.puta( 'myparam', 1000 ). For JSON file format
    to work, the values may not be arrays. So the pair puta() / geta()
    handles arrays of type numpy.ndarray by transparently converting such
    data into lists and back.'''
        tag = 'ndarray'
        if self.isstring( key ):
            if type( value ).__name__ == tag:
                # NOTICE then type( value ).__module__ is 'numpy'
                # encode ndarray as list
                valstor = [tag]
                vallist = value.tolist() # List of numbers
                valstor.append( vallist ) # The list as 2nd element of main list
            else:
                valstor = value
            self._dd['env'][key] = valstor
            return False # without error
        else:
            self.perror('Parameter key is no string. Quote, please: "key"!')
            return True
    #END puta() of Patt

    def geta( self, key ):
        '''Get an attribute value from the dictionary Patt._dd['env']
    Call: value = q.geta('wavelength'). If the attribute key is
    undefined, the method will raise a NameError.'''
        result = None
        if self.isstring(key):
            if key in self._dd['env']:
                result = self._dd['env'][key]
                if isinstance( result, list ) and (result[0] == 'ndarray'):
                    # Decode to numpy.ndarry
                    result = result[1]
                    result = np.asarray( result )
            else:
                self.perror('Key "%s" does not exist'%key)
            return result
        else:
            self.perror("Key is not a string -- quote it: 'name'!")
    #END geta() of Patt

    def hasa( self, key=None ):
        """Is key in the pattern attribute dictionary self._dd['env']?"""
        if key is None:
            key = ''
        if not self.isstring( key ):
            self.perror("Key is not a string -- quote it: 'name'!")
            return None
        if len(key.strip()) < 1:
            # Dump the whole attribute dictionary to the screen
            print('Attributes and their values:')
            attr = self._dd['env']
            for k in attr:
                print('%s  =  %s'%(k,attr[k]))
            return None
        return key in self._dd['env']
    #END hasa() of Patt

    def dela( self, key ):
        '''Delete the entry of key in the attribute dictionary of the pattern'''
        if not self.isstring(key, string):
            self.perror("Key is not a string -- quote it: 'name'!")
        if key in self._dd['env']:
            del(self._dd['env'][key])
    #END dela() of Patt
# END -------------Basic attribute management-----------------------------------------

# START +++++++++++Setters and Getters of damageable pattern parameters+++++++++++++++
    def _setvaxvars( self, vnlist ):
        '''Setter method for self._dd['env']['_vaxvars'].
It is related to the property vaxvars. Typical call:
        img.vaxvars = [ "$x$", "$y$", "$I$" ]
    Sets the axes-variable vector. Its components are TeX strings with curly
    braces escaped by double-curly braces (because matplotlib prints single
    curly braces directly). The last component describes the function itself.
    The components before describe the independent variables over which the
    function is defined. The vector must have at least 3 components. A
    definition in terms of TeX strings results in nicely labelled plots.

    TeX convention:
        # NOTICE upright characters inside a "dollar clause: \mbox{{nm}}
        #        bold vectors: \mathbf{{}} TODO: Special treatment for "\"?
    The attribute is accessed via setter/getter, because then the program
    can check direct user input.'''
        if not isinstance( vnlist, list ):
            self.perror("List: Use brackets [] to enclose parameters")
            return True
        if len( vnlist ) < 3:
            self.perror("List too short. Need 3 items")
            return True
        # Are all components strings?
        if not all(self.isstring(item) for item in vnlist):
            self.perror("All items in the list must be strings. Quote, please!")
            return True
        # Is information in every string?
        if not all( len(item.strip() ) > 0 for item in vnlist ):
            self.perror("Strings must not be empty")
            return True
        # Remove surrounding white space before storage
        vn = [item.strip() for item in vnlist]
        self.puta('_vaxvars', vn )
    #END _setvaxvars() of Patt

    def _getvaxvars( self ):
        '''Getter method for the axis labels which will be related to the
    property vaxvars (vector of axis variable names)
    Typical call:
        print(a.vaxvars)

    Get the list of variable identifier strings for the pattern. The vector
    must have at least 3 components of type string (horizontal variable,
    vertical variable, designation of the function itself)
    e.g.: [ "$x$", "$y$", "$I$"] or
          ["$s_{{12}}$", "$s_3$", "$z$"]'''
        return self.geta('_vaxvars')
    #END _getvaxvars() of Patt

    def _setvaxunits( self, vulist ):
        """Setter method for the attribute vaxunits (vector of axis unit names).
    It is associated to the property vaxunits.
    Typical call via property:
        img.setvaxunits = ["nm^{{-1}}", "nm^{{-1}}", "a.u."]
    or call via the direct method:
        img._setvaxunits( ["nm^{{-1}}", "nm^{{-1}}", "a.u."] )
    ... and this is equivalent to direct but unchecked setting like
        a.puta( '_vaxunits', ["nm^{{-1}}", "nm^{{-1}}", "a.u."] )
        a._dd['env']['_vaxunits'] = ["nm^{{-1}}", "nm^{{-1}}", "a.u."]

    Sets the variable-unit vector for the pattern in (at least) 3 axes.
    Variable units are, again, TeX strings. The last component describes
    the units of the function itself, e.g.
        'a.u.'            for arbitrary units or
        'e.u. nm^{{-3}}'  for electron-units per nm to the third.
    The components before the last describe the n-th, (n-1)st, ...
    component of the independent variables. For 2D patterns this means
    that the 2nd component describes the vertical axis (usually y) and
    the first component describes the horizontal or x-axis. This concept
    can easily be extended to functions of more than 2 parameters.

    By the setter the program can check the user input for correctness."""

        if not isinstance( vulist, list ):
            self.perror("Input must be a list: Use brackets [] to enclose parameters")
            return True # with error
        if len( vulist ) < 3:
            self.perror("List too short. Need 3 items")
            return True
        # Are all components strings?
        if not all(self.isstring(item) for item in vulist):
            self.perror("All items in the list must be strings. Quote, please!")
            return True
        # Is information in every string?
        if not all( len(item.strip() ) > 0 for item in vulist ):
            self.perror("Strings must not be empty")
            return True
        # Remove surrounding white space before storage
        vu = [item.strip() for item in vulist]
        self.puta('_vaxunits', vu)
    #END _setvaxunits() of Patt


    def _getvaxunits( self ):
        """The getter method for the axunits. Result is a list of strings"""
        value = self.geta('_vaxunits')
        return value
    #END _getvaxvars() of Patt


    def _getvaxras( self ):
        '''Getter method for the pixel sizes and function-value restriction
    vector. All components are numbers. The last component defines, if
    the function is permitted to have negative values. The components
    before that define the extensions of each pixel (voxel, supervoxel)
    in each of the directions over which the function is defined.'''
        return self.geta('_vaxras')
    #END _getvaxras() of Patt

    def _setvaxras( self, vdlist ):
        '''This setter method is related to the property vaxras!

          instance._dd['env']['_vaxras'] = [ 0.172, 0.172, 1 ]

        Sets the boxlengths of the pixels and in the last component a flag
        that indicates if the function should only take positive values
        like a scattering intensity (value > 0, e.g. 1).

        The reason for having this setter is controlling
        the user input for consistent parameters.'''
        if not isinstance( vdlist, list ):
            self.perror("List: Use brackets [] to enclose parameters")
            return True
        if len( vdlist ) < 3:
            self.perror("List too short. Need 3 items")
            return True
        # Are all components numerical?
        if not all(isinstance(item, (float,long,int)) for item in vdlist):
            self.perror("All increments must be numerical")
            return True
        # Check if genuine increments are positive
        genuine = vdlist[0:len(vdlist)-1] # Take off last element - That's Python
        if not all( item > 0 for item in genuine ):
            self.perror("Values (except the last one) must be positive")
            return True
        self.puta('_vaxras', vdlist )
    # END Patt._setvaxras

    def _setthemap( self, themap ):
        '''Check and set the mapp only'''

        try:
            dims = themap.ndim
        except:
            self.perror("Pattern is no numpy matrix")
            return True
        if ( dims != 2):
            self.perror("Pattern is not two-dimensional")
            return True
        if ( themap.size < 4 ):
            self.perror("Pattern has less than 4 pixels")
            return True
        if (themap.dtype != np.dtype('float32')):
            try:
                themap = np.float32( themap )
            except:
                self.perror("Conversion from list to np.float32 failed")
                return True

        # setup the map entry in the central data dictionary
        # and fill it with the supplied matrix which defines
        # the scattering pattern
        self._dd['map'] = themap
        # Enter width and height in the attribute dictionary, as well
        h, w = themap.shape
        self.puta('width', w)
        self.puta('height', h)
    #END _setthemap() of Patt

    def _getthemap( self ):
        """Getter method related to the property Patt.mapp for easy
    access of the pattern map Patt._dd['map']

    The image map is a numpy matrix of dtype float32. For instance, its
    height and width parameters are retrievable like this
        height, width = Patt.mapp.shape
    or via getters for easy access:
    Patt.height    Patt.width"""
        return self._dd['map']
    #END _getthemap() of Patt
# END ------------Setters and Getters of damageable pattern parameters---------------

# START ++++++++++Getters for increased usability in interactive python++++++++++++++
    def _mapheight( self ):
        """Getter for the map height"""
        h, w = self._dd['map'].shape
        return h
    #END _mapheight of Patt


    def _mapwidth( self ):
        """Getter for the map width"""
        h, w = self._dd['map'].shape
        return w
    #END _mapwidth of Patt

# END ------------Getters for increased usability in interactive python--------------

# START ++++++++++Property definitions ++++++++++++++++++++++++++++++++++++++++++++++
    mapp      = property(_getthemap, _setthemap) # MAP of the Pattern (numpy float32)
    height    = property(_mapheight)   # The height of mapp
    width     = property(_mapwidth)    # The width of mapp
    # "v"ectors of "ax"es descriptors
    vaxvars   = property(_getvaxvars, _setvaxvars) # Names of axes variables
    vaxunits  = property(_getvaxunits, _setvaxunits) # Units of the axes
    vaxras    = property(_getvaxras, _setvaxras) # Box lengths and pos flag

# START +++++++++Pattern I/O+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def save( self, filename, filformat=None ):
        '''Save the scattering pattern to a file. Supported file formats:
    json: The standard JSON format (JavaScript Object Notation). This is a
    .sjs  human-readable file format which is available for many other
          programming languages. It reads very fast, although the numpy-arrays
          must be converted to lists (2D arrays even to lists in a list).
    json-tricks: Extension of the json format so that it understands many
    .sjx  python data types more than the original. Still cross-platform, but
          no longer cross-language. It saves and loads slowly.
    pickle: The genuine python object storer and retriever. Dangerous,
          because it can contain malicious code. Fast.
    TODO  pgm for simple output of integers (interface to PoVRay and gimp)

    Call: p.save('test')  # Save in json format to 'test.sjs'
         p.save('test.spp') save in pickle format
         p.save('test.sjx') save in json-tricks format
         p.save('test', 'pickle') save to 'test' in pickle format
         p.save('test.spp', 'sjs') save in json format disguised as pickle'''

        supported  = ['json', 'jsonx', 'pickle']
        extensions = ['sjs', 'sjx', 'spp']
        if not self.isstring(filename):
            self.perror( 'filename is no string. Quote, please!' )
            return True
        # Is information in every string?
        if not len(filename.strip() ) > 0:
            self.perror("filename is an empty string")
            return True
        fn = filename
        ff = filformat
        if ff != None:
            if ff not in supported:
                self.perror('Unsupported file format: "%s"'%ff)
                return True
        else:
            # Append default extension if no extension present
            if fn.rfind('.') == -1:
                fn = fn+'.'+extensions[0]
            # Try to choose method by file extension
            ( mybody, sep, myext ) = fn.rpartition('.')
            if myext in extensions:
                ff = supported[extensions.index(myext)]
            else:
                # Choose default method
                ff = supported[0]

        if ff == 'json': # in the cross-language, primitive mode
            # jst.dump( self._dd, open( fn, "wb" ), primitives=True )
            aslist = self.copy()
            aslist._dd['map'] = self.mapp.tolist()
            if six.PY2:
                json.dump( aslist._dd, open( fn, 'wb' ) )
            else:
                json.dump( aslist._dd, open( fn, 'w', encoding='utf-8' ) )
        elif ff == 'jsonx': # in the python "json-tricks" mode
            if six.PY2:
                jst.dump( self._dd, open( fn, 'wb' ) )
            else:
                jst.dump( self._dd, open( fn, 'w', encoding='utf-8' ) )
        elif ff == 'pickle':
            pickle.dump( self._dd, open( fn, "wb" ) )
        print('Pattern saved to "%s" in format %s'%(fn,ff))
    #END save() of Patt

    def imaload( self, filename ):
        '''Almut Stribeck's compatibility routine. It loads patterns stored
    "in the stone age" by her PV-WAVE program collection. It is called by
    Patt.load(), but can be used standalone, as well.

    imaload loads a Patt from an ima file written by PV-WAVE using
    the XDR format (cross-platform data representation) for binary data as
    defined by SUN (the constructors of the famous graphical workstation
    of the 1990ties). XDR writes everything in big-endian notation
    (motorola processors), even if the machine works with little-endian
    numbers (intel processors). Moreover, XDR resolves the problem with
    differnt definitions of exponents and mantissae of floating point numbers.
    The method relies on python's xdrlib'''
        f = open( filename ).read()
        data = xdrlib.Unpacker(f)
        # Everything is Big-Endian '>l' (SUN!)

        #** Image width and height
        # We only need the first item of the tuples therfore the ","
        # Image width is a 4-byte signed integer ("long")
        imwidth,  = struct.unpack('>l', data.unpack_fstring(4))
        imheight, = struct.unpack('>l', data.unpack_fstring(4))

        #** Length of the original filename from the experiment
        titlelen, = struct.unpack('>l', data.unpack_fstring(4))
        #** Length of the last filename of the processed data
        fnlen, = struct.unpack('>l', data.unpack_fstring(4))

        #** Load the image map into a numpy map
        items = imwidth*imheight
        # linear for reading
        mapp = np.empty(shape=(items,1), dtype=np.float32)
        for i in xrange(items):
            mapp[i] = data.unpack_float()
        mapp = mapp.reshape((imwidth,imheight))

        # Image TITLE
        dummy = data.unpack_fstring(4) #Discard 1st of double string length
        title = data.unpack_string()
        # Image FILENAME
        dummy = data.unpack_fstring(4) #Discard 1st of double string length
        fn = data.unpack_string()
        # PIXEL SIZEs in horizontal and vertical direction are floats
        xdel = data.unpack_float()
        ydel = data.unpack_float()
        # CENTER of pattern in pixels. pv-wave reserves 4 bytes for each integer
        xcen, = struct.unpack('>l', data.unpack_fstring(4))
        ycen, = struct.unpack('>l', data.unpack_fstring(4))

        #** The timestamp:
        # YEAR: string of 4 bytes
        year, = struct.unpack('>l', data.unpack_fstring(4))
        # overread a flag that is 1 in big endian. It simply says:
        # The next item is 1 byte, but patched with zeros
        dummy = data.unpack_fstring(4)
        # MONTH: so we read the data in LITTLE-endian to catch the
        # padding zeros as well and to keep the 4-byte-measure
        month, = struct.unpack('<l', data.unpack_fstring(4))
        # DAY: same procedure
        dummy = data.unpack_fstring(4)
        day, = struct.unpack('<l', data.unpack_fstring(4))
        # HOUR
        dummy = data.unpack_fstring(4)
        hour, = struct.unpack('<l', data.unpack_fstring(4))
        # MINUTE
        dummy = data.unpack_fstring(4)
        minute, = struct.unpack('<l', data.unpack_fstring(4))
        # SECOND
        second = data.unpack_float()
        # The same date/time now as the julian number
        julian = data.unpack_double()
        # Catch the flag byte that is only good to tell pv-wave
        # to recompile julian from the other data
        dummy = data.unpack_fstring(4)
        flag, = struct.unpack('<l', data.unpack_fstring(4))
        #** The former "experimental parameters" of the pattern
        exppar  = data.unpack_float()
        exppar2 = data.unpack_float()
        exppar3 = data.unpack_float()

        # Make a new instance for the image
        # clean self
        self._makepatt( mapp )
        self.puta('title',title) # Where it came from
        self.puta('recorded',(year,month,day,hour,minute,second))
        self.puta('exppar',(exppar,exppar2,exppar3))
        self.puta('center', (xcen, ycen)) # The center of the pattern
        self.puta('fn',fn) # The file-name body
        # Default
        vaxvars = ["$s_{{12}}$","$s_3$","$I$"]
        vaxunits = ["nm^{{-1}}","nm^{{-1}}","a.u."]
        vaxras  = [xdel,ydel,1] # restricted to positive values
        self.puta('function', 'intensity')

        # Extract designator
        # We only consider the common file designators
        # '_gsf', '_CDF' -- anything else is "intensity"
        (mybody, sep, myext) = filename.rpartition('.')
        (titstamp, sep, desi) = mybody.rpartition('_')
        # Standard function descriptor for evaluated
        # scattering patterns in reciprocal space
        if desi == 'CDF':
            vaxvars = ["$r_{{12}}$","$r_3$","$z$"]
            vaxunits = ["nm","nm","a.u."]
            vaxras[2] = 0 # may take negative values
            self.puta('function','cdf')
        elif desi == 'gsf':
            vaxvars = ["$s_{{12}}$","$s_3$","$G$"]
            vaxras[2] = 0 # may take negative values
            self.puta('function','interference_function')

        # ima-masks always carry the substring "mask" in their name
        if filename.find('mask') != -1:
            vaxvars = ["$x$","$y$","$True$"]
            vaxunits = ["pix","pix"," "]
            vaxras = [1.0,1.0,0] # restricted to positive values
            self.puta('function','mask')

        self._setvaxvars( vaxvars )
        self._setvaxunits( vaxunits )
        self._setvaxras( vaxras )
    #END imaload() of Patt

    def load( self, filename, filformat=None ):
        '''load( filename [, format] )
    Fill the Patt from the data stored in a file

    Usage: import sapatt as sp
           r=sp.Patt()          # make an empty pattern
           r.load('test.spp')   # fill it from test.spp in pickle format
           r.load('a6p81c1_harm.ima') # ... from a file written by pv-wave'''

        supported  = ['json', 'jsonx', 'pickle', 'ima']
        extensions = ['sjs', 'sjx', 'spp', 'ima']

        if not self.isstring(filename):
            self.perror( "filename is no string. Quote, please!" )
            return True
        # Is information in every string?
        if not len(filename.strip() ) > 0:
            self.perror("filename is an empty string")
            return True
        fn = filename

        ff = filformat
        if ff != None:
            if ff not in supported:
                self.perror('Unsupported file format: "%s"'%ff)
                return True
        else:
            # Append default extension if no extension present
            if fn.rfind('.') == -1:
                fn = fn+'.'+extensions[0]
            # Try to choose method by file extension
            ( mybody, sep, myext ) = fn.rpartition('.')
            if myext in extensions:
                ff = supported[extensions.index(myext)]
            else:
                # Choose default method
                ff = supported[0]

        if not os.path.isfile(fn):
            print('File %s was not found or is unreadable' %fn)
            return True

        # .. and try to fill it
        if ff == 'json': # in the cross-language, primitive mode
            #self._dd = jst.load( open( fn, "rb" ), primitives=True )
            self._dd = json.load( open( fn, "rb" ) )
            # Convert mapp from a (primitive) json list to ndarray
            self._dd['map'] = np.asarray( self._dd['map'] )
        elif ff == 'jsonx': # in the python "json-tricks" mode
            if six.PY2:
                self._dd = jst.load( open( fn, "rb" ) )
            else:
                self._dd = jst.load( open( fn, "r", encoding='utf-8' ) )
        if ff == 'pickle':
            self._dd = pickle.load( open( fn, "rb" ) )
        elif ff == 'ima':
            self.imaload( fn )
        # Remember the name of the file from which the pattern has been read.
        self.puta('fromfile',fn)
        print('Pattern loaded from "%s" in format: %s'%(fn,ff))
    #End load() of Patt

    def rot90( self, n=None ):
        '''Rotate the pattern by n*90° in positive direction. If n is not
        given, the pattern is rotated once by 90°. Because the pattern
        does not know how rotation affects user-defined attributes, it only
        adapts the hard-coded attributes (width, height,center).
        If not yet present, the attribute "nrot90". If this attribute is
        defined, it holds the sum of all rot90 operations which have been
        performed on it-- of course modulo 4.'''
        rattr = "nrot90"
        cattr = "center"
        if n == None:
            n=1
        n = int(n) % 4
        if n == 0:
            return # Nothing to do
        self.mapp = np.rot90(self.mapp,n)
        if self.hasa(rattr):
            nsum = (self.geta(rattr) + n) % 4
        else:
            nsum = n
        self.puta(rattr,nsum)
        if n != 2:
            # Flip width and height *attributes*
            # self.width and self.height are retrieved directly
            # from the matrix.
            self.puta("width", self.width)
            self.puta("height", self.height)
            vars = copy.copy(self.vaxvars)
            # Flip x- and y-axis designators
            self.vaxvars[0], self.vaxvars[1] = vars[1], vars[0]
            vars = copy.copy(self.vaxras)
            self.vaxras[0], self.vaxras[1]   = vars[1], vars[0]
            if self.hasa('_vaxunits'):
                vars = copy.copy(self.vaxunits)
                self.vaxunits[0], self.vaxunits[1] = vars[1], vars[0]

            if self.hasa(cattr):
                co = self.geta(cattr)
                cn = co
                cn[0], cn[1] = co[1], co[0]
                self.puta(cattr,cn)
        pass
    #END rot90() of Patt

    def _scale_log_pos( self, ulev=None ):
        '''Returns a logarithmically scaled version of the positive face of
    the pattern map on a logarithmic scale that extends from the smallest
    positive value to ulev. Mainly for internal use.
        Usage:
             alog = a._scale_log_pos( 1000 )'''

        # A copy for the result
        res = self.copy()

        top = np.max(self.mapp)
        # Set all values that are smaller than the smallest positive number
        # to the smallest positive number in the map
        scratch = copy.copy(self.mapp)
        scratch[scratch <= 0] = top # push up to top: zeroes and values below
        posmin = np.min(scratch)

        # Small positive value (no arithmetic error), but almost black
        res.mapp[res.mapp <= 0] = posmin

        # Normalize the positive range in order to obtain a nice
        #    logarithmic gradient for color display
        trumax = np.max(res.mapp)
        if ulev is not None: # The user supplies an own upper level?
            if ulev > 0.0:
                top = ulev # use it!
        res.mapp = 100.0*res.mapp/top # is 100.0 at top
        # Switch to logarithmic scale
        scratch = np.log( res.mapp )
        toplog = np.log(100.0) # Top value in scratch
        res.mapp = top*scratch/toplog
        return res
    # End _scale_log_pos() of Patt


    def tolog( self, *args ):
        '''scales a pattern to logarithmic "intensity". The method works as
    well on patterns that have zero or negative values (both branches will be
    put on each a logarithmic scale)
        Usage:
               b = a.tolog()
               b = a.tolog("stretch")
    If called with the keyword stretch, the negative and the positive branch
    of a function will be scaled independently from each other to full scale
    (this is only interesting in some weird cases)'''
        # Make a copy of the pattern that takes the result
        res = self.copy()
        pm = np.max( self.mapp )
        nm = -np.min( self.mapp )
        if (pm != nm):  # exclude trivial case
            if (len(args) > 0) and (args[0] != "stretch"):
                if (pm < nm):
                    pm = nm
                nm = pm
        upimg = res._scale_log_pos( pm )
        imginv = self.copy()
        imginv.mapp[...] = -self.mapp[...]
        dnimg = imginv._scale_log_pos( nm )
        dnimg.mapp[...] = -dnimg.mapp[...]
        # because upimg is 0 where dnimg is not and vice versa
        #   the following line works
        res.mapp[...] = pm*(upimg.mapp[...] + dnimg.mapp[...])
        # A general way to store absolute values requires
        # two lines of code, as well:
        # x = np.absolute(res.mapp)
        # res.mapp[...] = x[...]
        return res
    #END tolog() of Patt


    def pix_to_sci( self, ix, iy ):
        '''Transform screen pixel coordinates ix, iy to units of the
        scientific coordinate system. The coordinates may be float
        numbers, as is usual with matplotlib return values.'''
        xi, yi = ix, self.height - iy # NOTICE from image to math convention
        xi, yi = int(xi+0.5), int(yi+0.5)
        cen = self.geta('center')
        xi, yi = xi - cen[0], yi - cen[0] # shift to center

        # The origin of the pattern is in the lower left corner of the center
        # pixel. A well-centered pattern of 512 x 512 pixels has its center
        # in (256,256). By this definition a pattern can have 4 identical
        # quadrants. Moreover, there is no pixel centered about the origin
        # and no division by zero can occur in working with the coordinates
        # of pixels.
        # NOTICE For Fourier transforms a quadrant must be picked and
        #        shifted back by interpolation.
        xres = (xi+0.5)*self.vaxras[0] # via getter
        yres = (yi+0.5)*self.vaxras[1]
        return (xres, yres)
    #END pix_to_sci() of Patt

    def neg_and_pos( self ):
        '''Returns True, if the pattern has both positive and
        negative values'''
        top = np.max(self.mapp)
        bot = np.min(self.mapp)
        return (top > 0.0) and (bot < 0.0)
        pass
    #END neg_and_pos() of Patt

    def isamask( self ):
        '''Returns True if the pattern is (almost) a binary mask'''
        top = np.max(self.mapp)
        bot = np.min(self.mapp)
        if top == bot:
            # Degenerate function
            print('This is a constant mapp!')
            return False
        m = np.ma.masked_equal(self.mapp,top)
        tops = m.mask.astype(int).sum()
        m = np.ma.masked_equal(self.mapp,bot)
        botts = m.mask.astype(int).sum()
        h, w = self.mapp.shape
        pixels = 1.0*h*w # may run out of integers
        return ((tops + botts)/pixels) > 0.98
    #END isamask() of Patt


# +++++++++++++Morphological operators+++++++++++++++++++++++

    def erosion( self, size ):
        '''Erode the pattern in place'''
        self.mapp = ndimage.grey_erosion(self.mapp,size)
        # NOTICE accounting for accumulated erosion is a bit cumbersome
        # self.puta('eroded',size)
        pass
    #END erosion() of Patt

    def median( self, size ):
        '''median filter the pattern in place'''
        self.mapp = ndimage.median_filter(self.mapp,size)
        self.puta('median_filtered',size)
        pass
    #END median() of Patt

    def dilation( self, size ):
        '''Dilate the pattern in place'''
        self.mapp = ndimage.grey_dilation(self.mapp,size)
        pass
    #END dilation() of Patt

    def opening( self, size ):
        '''Open the pattern in place'''
        self.mapp = ndimage.grey_opening(self.mapp,size)
        pass
    #END opening() of Patt

    def closing( self, size ):
        '''Close the pattern in place'''
        self.mapp = ndimage.grey_closing(self.mapp,size)
        return
    #END closing() of Patt

#----------------- Morphological operators----------------------------

# ++++++++++++++++ Display methods +++++++++++++++++++++++++++++++++++


    def __prepshow( self, draw ):
        '''Prepare the draw map for showing according to the user's wish
        Here the map i prepared in such a way that a value of 1.0 shall be
        mapped on the upper edge of the color map (CLUT). In pyplot this is
        unnnecessary, because it always maps the z-range of the function on
        one loop of the color map. If this is not accepted, the programmer
        has the option to set vmin= and vmax= . This one of the many
        features which are not documented in the description of the library.
        How I love the 1 m wide shelf of reference guides of PV-WAVE!
        So for now I will not redesign this module, but simply always set
        vmax=1.0'''
        ulev = copy.copy(self.drawulev) # May change with the _freshi's

        top = np.max(draw.mapp)
        bot = np.min(draw.mapp)

        if draw.isamask():
            # A general 2-value function
            # Map it to [0,1] for visualization
            draw.mapp[...] = (draw.mapp[...] - bot) /(top - bot)
            # Disable all "view modes"
            return 1.0, draw

        if self.drawneg:
            # Turn mapp upside down, if ...
            # Check, how many negative values there are
            m = np.ma.masked_less(draw.mapp,0) # Search negative values
            howmany = m.mask.astype(int).sum()
            if howmany > 20: # ... there are enough negative values
                draw.mapp[...] = -draw.mapp[...]

        if self.drawupface:
            # remove negative and zero values for the plot
                scratch = copy.copy(draw.mapp)
                scratch[scratch <= 0] = top
                posmin = np.min(scratch)
                #  Use posmin to mask a dark region
                draw.mapp[draw.mapp <= 0.0] = posmin

        if self.drawlog:
            # scale function's upward face logarithmically
            upimg = draw._scale_log_pos( ulev )
            # scale function's downward face logarithmically
            if np.min(draw.mapp) < 0.0:
                # only if it has relevant data
                draw.mapp = -draw.mapp # Flip bottom up
                dnimg = draw._scale_log_pos( ulev )
                draw.mapp = upimg.mapp + dnimg.mapp
            else:
                draw.mapp = upimg.mapp

        top = np.max(draw.mapp)
        bot = np.min(draw.mapp)

        if top < 0.0:
            # Caution: The function is negative everywhere
            #          Push it up so that bot becomes 0.0
            draw.mapp = draw.mapp - bot
            top = top - bot
            bot = 0.0

        ulevel = ulev
        if not ulevel:
            # Wish for auto scaling
            if top == bot:
                ulevel = 1.0 # trivial mapp
                draw.mapp = draw.mapp - bot # pull it black
                bot, top  = 0.0, 0.0
            else:
                ulevel = top

        # color gradient in neg direction shall be same as in pos
        draw.mapp[...] = abs(draw.mapp[...])
        # apply modulo ulevel
        draw.mapp = np.fmod(draw.mapp,ulevel)
        # matplotlib CLUT end is at 1.0 (not at 255)
        draw.mapp = draw.mapp/ulevel
        #print('2: %10.5e %10.5e'%(np.min(draw.mapp),np.max(draw.mapp)))
        return ulevel, draw
    #END __prepshow() of Patt


    def showi(self, piaw, wintit=None, ulev=None, clut=None, negative=False, logscale=False):
        '''Show a 2D scattering pattern in pseudo color with the possibility
        to modify it interactively.
            Usage: import sapatt as sp
                   import sapiawin as piaw
                   self.piawinds = {}
                   pia = piaw( self.piawinds, verbose=True )
                   a = sp.Patt()
                   a.load('mypattern')
                   a.showi(pia, wintit='mypattern', ulev=0.3, clut='gnuplot')'''
        if ulev:
            ulev = abs( ulev ) # must be positive
        if clut is None:
            clut = 'CMRmap' # color lookup table
        # accessed by self._freshi:
        self.drawpia  = piaw
        self.drawtit  = wintit
        self.drawulev = copy.copy(ulev)
        self.drawclut = clut
        self.drawneg  = negative
        self.drawlog  = logscale
        self.drawupface = False # We always start with a complete view

        # Make a work copy of the pattern
        draw = self.copy()

        #NOTICE TODO __prepshow() has been written with a wrong
        # understanding of the matplotlib.pyplot.imshow() method.
        # When I wrote it, i understood that it would work in the
        # fashion of PV-WAVE, where the top value (here: 1 in
        # WAVE: 255) was mapped on the top of the CLUT. Here the
        # CLUT is mapped dynamically in side imshow, if vmax is not
        # set.
        ulevel, draw = self.__prepshow( draw )

        # Without interactivity we could simply pop up something:
        # plt.imshow( draw.mapp, cmap=clut )
        # plt.show

        # keep info for _freshi
        #TODO with clutop=1.0 we set the top of the CLUT to the value which
        #     we want to have. This changes, if __prepshow() is refurbished.
        self.drawwin = piaw.popup( draw.mapp, clut=clut, clutop=1.0, wintit=wintit,
                                  onwsel=self._onwaction,onwkey=self._onwview )
        if not ulev:
            print('[%s] Auto scaling. ulev: %10.3e'%(self.drawwin,ulevel))
    #END showi() of Patt


    def _freshi( self ):
        '''Fresh up the image after an interactive modification'''
        # Make a work copy the pattern
        draw = self.copy()

        ulevel, draw = self.__prepshow( draw )
        if not self.drawulev:
            # Automated ulevel may have changed
            print('[%s] Auto scaling. ulev now: %10.3e'%(self.drawwin,ulevel))

        pia    = self.drawpia
        winnum = self.drawwin
        wintit = self.drawtit
        pia._freshup( draw.mapp, winnum, wintit=wintit )
    #END _freshi() of Patt


    def roigon2mask( self, oneroi ):
        '''Takes the polygonal oneroi and generates a mask which is True inside
        the polygon and False outside. The x-values of the polygon must be in
        oneroi[0] and the y-values in oneroi[1]. The first and the last point
        of the polygon must be identical (closed polygon).'''
        xgon = oneroi[0]
        ygon = oneroi[1]
        # Convert the polygon from Line2D convention to vertices
        # NOTICE must embed in list(), because pure zip returns an iterobj
        gonverts = list(zip(xgon,ygon))
        gonpath = Path(gonverts) # Method from matplotlib.path

        # Generate the mesh that makes the pixels
        w, h = self.width, self.height # extension of the mapp in pixels
        x, y = np.meshgrid( np.arange(w), np.arange(h) )
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T

        # Generate the mask with the pixels inside the polygon
        # set to True
        mask = gonpath.contains_points(points)
        mask = mask.reshape((h,w))
        return mask
    #END roigon2mask() of Patt


    def _onwview( self, event, windict ):
        '''Handler of keyboard keys which act on the pattern.
        These may be view changes, but also actions which are
        not related to the window, like saving'''

        winnum = event.canvas.figure.number

        #** Actions not related to the window
        if event.key.lower() == 'alt+s':
            # save the pattern
            if self.hasa('fromfile'):
                fnbase = self.geta('fromfile')
                if fnbase.rfind('.') != -1:
                    ( body, sep, ext ) = fnbase.rpartition('.')
                    fnbase = body
            elif self.hasa('fn'):
                fnbase = self.geta('fn')
            else:
                # actual state
                fnbase = 'as'
                # might append a time stamp
            # prefix for easy finding, suffix for the cross-platform
            # and fast standard json format
            savename = '__%s.sjs'%fnbase # NOTICE in PY2 not "str", but "unicode" type
            print('[%s] Saving the actual pattern to "%s"'%(winnum,savename))
            self.save(savename)

        #** Actions which are related to the open window.
        # This window's status
        winstat = windict[winnum]
        thefig  = winstat['fg']
        theaxes = winstat['ax']
        if isinstance(theaxes, tuple):
            theaxes = theaxes[0]
        thecids = winstat['ci'] # Dictionary of connector ID's

        if event.key in ['?','f1']:
            # Toggle the visibility of the hint
            plthint = winstat['hi']
            visible = plthint.get_visible()
            # print('[%s] hp: "%s" visi: %s'%(winnum,plthint,visible))
            plthint.set_visible( not visible )
            plt.show() # Display change!
        elif event.key.lower() == 'o':
            # Toggle logarithmic scaling of the image
            self.drawlog = not self.drawlog
            self._freshi()
        elif event.key == '-':
            if self.neg_and_pos():
                # Toggle turning function bottom up
                self.drawneg = not self.drawneg
                # And indicate bottom up status
                pltbohi = winstat['bo']
                pltbohi.set_visible( self.drawneg )
                self._freshi()
        elif event.key == '+':
            # Toggle showing only the upward looking face
            self.drawupface = not self.drawupface
            self._freshi()
        pass
    #END _onwview() of Patt

    def _onwaction( self, event ):
        '''Handler of selections made in the open window'''
        winnum = event.winnum   # The window number
        pr     = event.polyroi  # The polygonal roi
        flavor = event.flavor   # line, box, bar, circle, ring
        task   = event.keytask  # What to do with the roi
        x1, y1, x2, y2  = event.vector # Mind the "classical" arrangement!

        segs = task.split('+')
        if len( segs ) == 2:
            modifier, charac = segs
        else:
            modifier, charac = '', task
        down = charac.lower() == charac # was lower case character

        if flavor == 'line':
            modifier = '' # A line does not enclose any roi
            # Make the line coordinates available for further processing
            self.lastline = event.vector # (x1,y1,x2,y2)
        else:
            # Compute the roi
            roi = event.polyroi
            mask = self.roigon2mask( [roi[0],roi[1]] )
            # NOTICE The conversion to a mask of numbers is readily done by
            # maskint = mask.astype(int) # True -> 1, False -> 0
            if event.flavor == 'ring':
                # The mask of the bigger circle
                maskbig = self.roigon2mask([roi[2],roi[3]])
                # True between the two circles
                mask = np.logical_not( mask )
                mask = np.logical_and(mask, maskbig)
            omask = np.logical_not( mask ) # What is outside

        if modifier == '':
            if flavor == 'line':
                pixel = (int(x2+0.5), int(y2+0.5)) # look up the end point of the vector
                mess1 = '[%s] Line pointing at'%winnum
            else:
                # We have a mask and must search something
                mappt = copy.copy(self.mapp) # working mapp for testing
                if down:
                    prop = 'Minimum'
                    # Search the minimum inside the roi
                    push = np.max( mappt )
                    mappt[omask] = push # push everything outside up
                    pixel = np.unravel_index(np.argmin(mappt), mappt.shape)
                    # This always only returns the first tuple in which the
                    # minimum is reached.
                    # Other minimum values can be highlighted with
                    #allmin = self.mapp == self.mapp[pixel]
                    #howmany = allmin.astype(int).sum()
                else:
                    prop = 'Maximum'
                    # Search the maximum inside the roi
                    push = np.min( mappt )
                    mappt[omask] = push # push outside down
                    pixel = np.unravel_index(np.argmax(mappt), mappt.shape)
                mess1 = '[%s] %s inside %s found at'%(winnum,prop,flavor)
            x, y = pixel
            # WARNING Strange: The coordinates appeared strangely swapped, but now they
            #         behave normal.
            funval = self.mapp[x,y] # Function's value at the pixel pos
            print('%s (%i,%i) is %10.3e'%(mess1,x,y,funval))
            x, y = self.pix_to_sci(x, y) # Now in scientific coordinates
            reciprocal = ( self.vaxunits[0].find('{{-1}}') > -1 ) or ( self.vaxunits[1].find('{{-1}}') > -1 )
            if reciprocal:
                sx, sy = x, y
                s = np.sqrt(sx**2 + sy**2)
                u = np.float(1.0)
                r = 1.0/s
                rx, ry = 1.0/sx, 1.0/sy # Pattern geometry definition avoids div by zero
                print('  (s12,s3,s) = (%7.4f, %7.4f, %7.4f) [1/nm]'%(sx,sy,s))
            else:
                rx, ry = x, y
                r = np.sqrt(rx**2 + ry**2)
            print('  (r12,r3,r) = (%7.1f, %7.1f, %7.1f) [nm]'%(rx,ry,r))

            # Make the circle's center available for a method which asks to
            # define the center of this pattern by input of a circle.
            if flavor == 'circle':
                xc, yc = int(x1+0.5), int(y1+0.5)
                pixel = (xc, self.height - yc) # from image to mathematical notion
                self.lastcenter = pixel
            return # No modification, no _freshup - so return!

        # Modify the roi
        if down:
            push = np.min( self.mapp ) # Down to the minimum
            #print('Pushing down to the minimum %10.3e'%push)
        else:
            push = np.max( self.mapp ) # Up to the maximum
            #print('Pushing up to the maximum %10.3e'%push,'')

        if modifier == 'alt':
            # Modify the outside of the roi
            self.mapp[omask] = push
            #print(' the outside of the roi')
        elif modifier == 'ctrl':
            # Modify the roi itself
            self.mapp[mask] = push
            #print(' the roi')
        # Show the result
        self._freshi()
    #END _onwaction() of Patt

#END class Patt

