#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import numpy as np

'''Inject values from a (smoothed) attribute curve into the frames
taken during an experiment This curve must be in OP2 and must be
a function of the frame index. The script destroys OP0, where it
reads in the frame curves.
SYNTAX: @@attrinj <framefiles> <keyy>
    <framefiles>: Selector defining the file series into which the
                  values shall be injected
    <keyy>:  the key under which the values from the curve shall be
             stored in the curve dictionary
    EXAMPLEs:
             # The examples are related to a curve structure, in which
             # beam attenuation values (bat) are stored in the attribute
             # "bat", but these values are noisy. So they have been extracted
             # by attrext.py, then smoothed by a saskia command. Then this
             # python script stores the smoothed attribute values under 
             # the new key "bats":
             # to store the smoothed attributes into the (new) key bats
             @@attrinj,sy02_up*,bats
             # If one would simply overwrite into the old key "bat", which
             # is not recommended, one would call:
             @@attrinj,sy02_up*,bat
        '''

## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def injector():
    # *** Check the number of arguments with which the script has been called
    argus = len(sys.argv) - 1
    if argus != 2:
        return "script needs the 2 arguments <framefiles> and <keyy>"
    # Check presence and properties of the injector curve in OP2
    if self.ops[2].isempty():
        return "OP2 must contain the injector curve, but it is empty"
    injabscissa = self.ops[2].attribute('axis1')
    if  injabscissa != 'frame':
        return 'OP2: The injector curve is not a function of "frame", but of "%s"'%(injabscissa)
    # In the injector the frames are stored as floats. So they
    # must by converted to integers before matching is possible.
    ispan = self.ops[2].n
    injxi = np.round(self.ops[2].x[:ispan]).astype(int)
    injyi = self.ops[2].y[:ispan] # The y-values
    # *** Evaluate the arguments
    # Argument 1: the "all"-files
    alldef = sys.argv[1]

    # If the user has given no extension, then append '.dat'
    test = alldef.split('.')
    if len(test) < 2:
        alldef = alldef+'.dat'

    # Argument 2: The key under which the values shall be injected
    dictkey = sys.argv[2]

    # alphabetically sorted list of the "all"-files
    allfiles = sorted(glob.glob(alldef))
    if len(allfiles) < 1:
        return 'No files matching "%s"'%alldef

    scat = 0 # Index where the scattering frames are read
    actframe = self.ops[scat] # Operand where the frames are read

    for i, thefile in enumerate(allfiles):
        #self.poutput( thefile ) #NOTICE for testing
        err = self.ops[scat].readascii( thefile )
        if err != '':
            return err
        # May be that the user has erased some of the frames. Then "i" is
        # not the frame index. So let's extract the frame index from the
        # file name!
        body  = thefile.split('.')
        words = body[0].split('_')
        frame = words[-1]
        xval = int(frame) # lookup-abscissa value in the injector
        match = np.where( injxi == xval ) # Lookup in the injector
        midx = match[0][0] # Isolate the match index from the output of where
        if midx is None:
            return 'For frame %s there is no matching value in OP2. Injection failed.'%xval
        yval = injyi[midx] # This is the value to be injected
        actframe.attribute(dictkey,yval) # INJECT IT!
        # Write the curve back to the same file name. "True" enforces overwriting
        err = self.ops[scat].writeascii( thefile, True )
        if err != '':
            return err
        # Indicate progress
        self.poutput('%s: attribute env[%s]=%s has been set.'%(thefile,dictkey,yval))
        if err != '':
            return err
        pass # loop end
    pass
#END injector

# The main program of this script
error = injector()
if error:
    self.perror(error)
    # And exit without signaling the error status
banner()

