#!/usr/bin/env python
# coding=utf-8
import os
import sys
import numpy as np

'''mergecAT.py: "merging curves" with environmental data collected at the
synchrotron "A"LBA, BL11-NCD in an experiment, in which the temperature "T
"is varied. No further processing.

THIS IS THE FIRST STEP OF DATA EVALUATION, PROVIDED THE SCATTERING PATTERNS
ARE ALREADY CONVERTED INTO CURVES. SUCH PYTHON SCRIPTS MUST BE WRITTEN EVERY
TIME AFTER A NEW EXPERIMENTAL CAMPAIGN. EACH SCRIPT IS ADAPTED TO THE DATA
FORMAT OF THE RAW SCATTERING DATA, THE DATA FORMAT OF THE ENVIRONMENTAL DATA,
AND TO THE  SPECIAL ENVIRONMENTAL DATA THAT MUST BE CONSIDERED DURING THE
FURTHER PROCESSING OF THE DATA.

SYNTAX:  @@mergecAT,<expfile>,<outfileprefix>,"saxs"|"waxs","collbat"
             <expfile>: The experiment log file written by ALBA. It is made
                        from the components "experiment" (e.g. "Y023_up"),
                        "_run" (e.g. "_000") and the standard extension
                        "_ExptEnv.Alba". If not given, the standard extension
                        will be automatically appended.
             <outprefix>: defines the user's wish to name the curve files
                        which belong to this frame run. The curve files will
                        be named <outprefix>_xxxx.dat with xxxx being the
                        frame number. One might decide to name the files
                        "s"axs of sample "y02" in the temperature "_up"
                        frame run or branch.
             <detector>: must be the string "saxs" or the string "waxs" to
                        discriminate between the SAXS and the WAXS data.
             <collbat>: if the 3rd argument is the word "collbat", then
                        the attenuated beam intensity "bat" (normalized to
                        the primary beam intensity) is, additionally,
                        collected in OP2 -- for direct smoothing followed
                        by re-injection. One may store this curve manually
                        (e.g. for the processing of WAXS data or
                        documentation).
    WARNING In this version the raw curves are immediately condensed to
            half of the number of points which are stored in the raw curve.
Example: @@mergecAT,Y023_up_000,sy02_up,saxs
Continue with extraction of the noisy "beam attenuation data":
         @@attrext,sy02_up*,ay02_up_bat,bat
Then average / smooth the beam attenuation curve and store it in a new
curve (e.g.:read,ay02_up_bat&smp,0,0,1,1,2.0&write,ay02_up_bat_smo),
which is injected into the frameset as new attribute "bats".
In WAXD data it is important to remove the aluminum peaks in the
background and in the data curves. This will become tricky, if the sample
itself contains big crystals. For polymers a simple solution is, again,
smoothing. For the background try
     read,rayonix_bg_000_0000 &smp,0,0,1,2,200.0 &gra
For me this extreme smoothing parameter of 200 removed the metal peaks without
changing the background considerably.
For the sample curves we can be more cautious, because the aluminum peaks
to not stick out that much. Here a smoothing parameter of 10 to 20 worked
well:
     read,rayonix_YT10_000_0013 &smp,0,0,1,2,10.0
Tune after every beamtime!'''


def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def merger():
    '''The central merging procedure'''

    ### Constant(s)
    batkey = 'bat' # Key under which the PIN-diode data are stored

    ### Argument parsing
    #TODO expand argument list
    arglist = argexpand(4) # Guarantee 4 arguments
    ## (1) The ALBA log file to be processed
    logfile = arglist[0]
    # Search the file
    if not os.path.isfile(logfile):
        logfile = logfile+'_ExptEnv.alba' # Append default extension
    if not os.path.isfile(logfile):
        self.perror('Experiment log "%s" was not found or is unreadable'%logfile, traceback_war=False)
        return
    self.poutput('Experiment log: %s'%logfile)
    ## (2): file name prefix for output curve files
    fnpre = arglist[1]
    if fnpre == '':
        self.perror('2nd arg is missing (prefix outfiles)', traceback_war=False)
        return
    self.poutput('Prefix output files: %s'%fnpre)
    ## (3): process waxs or saxs data?
    if arglist[2].lower() == 'saxs':
        usewaxs = False
    elif arglist[2].lower() == 'waxs':
        usewaxs = True
    else:
        self.perror('3rd arg must be "saxs" or "waxs"', traceback_war=False)
    ## (4): Readings of the PIN diode shall already be collected in OP1
    collbat = arglist[3].lower() == 'collbat'
    ### End argument parsing

    # General initiation
    basket = {} # Dictionary to collect shared environmental data
    if collbat:
        self.ops[1].clean() # Prepare for collection of PIN-diode readings
        head = '# "bat" (attenuated beam data) from %s'%logfile
        self.ops[1].headers.append( head )
        self.ops[1].env = {'axis2' : batkey, 'axis1' : 'frame' }

    # LOGFILE scanning
    with open(logfile,'r') as lf:
        for i, line in enumerate(lf):
            if line[-1] == '\n': # All lines carry a line feed -- but the last line may not.
                line = line[:-1] # cut off line feed character
            if i == 0:
                # First line of .alba file: The time stamp
                ts = line[6:].strip() # Weekday removed
                tsls = self._normaltimestr( ts ) # Time stamp of log file (string)
                dicts = "{'TimeStamp' : '%s'}"%tsls
                basket.update(eval(dicts)) # Add to basket
            elif i == 3:
                # Sample/Run is in the 4th line
                elem = line.split(' ')
                tag = elem[1].strip()
                if tag != 'Prefix:':
                    self.perror('%s: Syntax error: Prefix not in line 4.'%logfile, traceback_war=False)
                    lf.close()
                    return
                tag = 'sample_run' # Rename this item
                value = elem[2].strip()
                value = value[:-1] # Remove trailing underscore
                # Make a one-phrase dictionary
                dicts ="{'%s' : '%s'}"%(tag, value)
                basket.update(eval(dicts)) # add to the basket
            elif i == 4:
                # Exposure is in the 5th line
                elem = line.split(' ')
                tag = elem[1].strip()
                if tag != 'exposure:':
                    self.perror('%s: Syntax error: Exposure not in line 5.'%logfile, traceback_war=False)
                    lf.close()
                    return
                tag = tag[:-1] # strip colon
                value = elem[2].strip()
                # Make a one-phrase dictionary
                dicts ="{'%s' : %s}"%(tag, value)
                basket.update(eval(dicts)) # add to the basket
            elif i == 5:
                # Frames is in the 6th line
                elem = line.split(' ')
                tag = elem[1].strip()
                if tag != 'frames:':
                    self.perror('%s: Syntax error: Frames not in line 6.'%logfile, traceback_war=False)
                    lf.close()
                    return
                # This is a general variable similar to timestamp. We keep it
                # locally for computations
                value = elem[2].strip()
                frames = eval( value )
            elif i == 7:
                # latency is in the 8th line
                elem = line.split(' ')
                tag = elem[1].strip()
                if tag != 'latency:':
                    self.perror('%s: Syntax error: Latency not in line 8.'%logfile, traceback_war=False)
                    lf.close()
                    return
                tag = tag[:-1] # strip colon
                value = elem[2].strip()
                # Put latency in the basket
                dicts ="{'%s' : %s}"%(tag, value)
                basket.update(eval(dicts))

                # General items for the basket: All the curves are intensities.
                # They are transformed in the s-system for convenient maths.
                gen={'axis2': '$I$', 'axis1': '$s$', 'units2': 'a.u', 'units1': 'nm$^{{-1}}$'}
                basket.update(gen)

                # Initiation of control variables before iteration of frames
                cycletime = basket['exposure'] + basket['latency']
                firstflash = basket['exposure']/2.0 # Time of first frame
                pass
            elif i > 9:
                # Frame log lines:
                # Here are the lines which describe the varying environment of each frame
                fdata = line
                chan = fdata.split('\t') # The channels are separated by tabs
                # Replace '.edf' (pointer to 2D data) by '.dat" (pointer to curve)
                sascurve = chan[0][:-4]+'.dat' # The name of the small-angle CURVE
                wascurve = chan[1][:-4]+'.dat' # The name of the wide-angle CURVE
                pin = eval(chan[11]) # Beam attenuation detector reading [a.u.]
                #NOTICE: ***At ALBA pin is already divided by exposure***
                BeamstopGateOpen = eval(chan[13]) # same as basket['exposure']
                # Intensity of the incident beam: Synchrotron current in mA
                beam = eval(chan[12])
                if beam <= 0:
                    beam = 163.0 # replace by ALBA standard WARNING
                T  = eval(chan[14]) # The temperature of the sample [°C]
                if (pin <= 0) or (beam <= 0):
                    # NOTICE pin is typically 2.4e-5
                    bat = 9.99 # Dummy value -- must inject from other run!
                    print('BEAMLINE SETUP SHORTCOMING! NO beam attenuation data!')
                else:
                    bat = pin/beam # PIN-diode intensity normalized to the synchrotron current
                # Put all the actual environmental parameters into the dictionary "special"
                special = eval("{'%s' : %s}"%('T', T))
                item = eval("{'%s' : %s}"%('beam', beam))
                special.update(item) # Add to special
                item = eval("{'%s' : %s}"%(batkey, bat))
                special.update(item) # Add to special
                if usewaxs:
                    item = eval("{'%s' : '%s'}"%('rawfile', wascurve))
                    readfn = wascurve # For easy access
                else:
                    item = eval("{'%s' : '%s'}"%('rawfile', sascurve))
                    readfn = sascurve # For easy access
                special.update(item) # Add to special
                ## Handle frame and elapsed time
                ## extract the frame indicator string
                frameis = readfn[-8:-4] # four digits
                actframe = int(frameis)
                item = eval("{'%s' : %s}"%('frame', actframe))
                special.update(item)
                # Computing elapsed for each frame will not stumble, if the user
                # deletes single lines from the log file
                elapsed = firstflash + actframe*cycletime
                item = eval("{'%s' : %s}"%('elapsed', elapsed)) # in seconds
                special.update(item)

                ### The "real" processing of the actual frame
                ## Read the curve data into OP1, but strip the comments
                ## NOTICE and condense to half of the points in the curve
                ##        We do not need 5800 data points
                self.poutput('Reading raw curve: %s'%readfn)
                _err = self.ops[0].readascii( readfn, True, True )
                if _err != '':
                    self.perror( _err, traceback_war=False)
                    lf.close()
                    return True
                ## Supply the environment data
                self.ops[0].env.update(basket)
                self.ops[0].env.update(special)
                ## The frame info in one line
                frameinfo = '# %s %s. Elapsed:%8.2f s. T: %6.1f °C'%(readfn,tsls,elapsed,special['T'])
                self.ops[0].headers.append( frameinfo )
                ## Normalize to incident intensity and exposure
                ## The 1.0E3 is nothing but an arbitrary balancing factor
                Norm = 1.0E3/(special['beam']*basket['exposure'])
                span = self.ops[0].n
                self.ops[0].y[:span] = self.ops[0].y[:span]*Norm
                ## convert from q-system to s-system
                q2s = 1.0/(2.0*np.pi)
                self.ops[0].x[:span] = self.ops[0].x[:span]*q2s
                ## No background correction. With respect to earlier treatment,
                ## where the PIN diode collected many counts, now background
                ## correction must be postponed until the bat-data are
                ## smoothed across the frames (cf. collbat).
                savfn = '%s_%s.dat'%(fnpre,frameis)
                self.poutput('Write curve to %s'%savfn)
                err = self.ops[0].writeascii( savfn, True ) # overwrite
                if err != '':
                    self.perror( err, traceback_war=False)
                    lf.close()
                    return
                ### Bookkeeping
                self.poutput( frameinfo )
                if collbat:
                    actpoint = self.ops[1].n
                    self.ops[1].x[actpoint] = actframe
                    self.ops[1].y[actpoint] = bat
                    self.ops[1].n = actpoint+1

    lf.close()
    self.ops[0].clean() # erase the scattering curve
    myname = os.path.basename(sys.argv[0])
    if usewaxs:
        self.poutput('{!r} has preprocessed WAXS curves'.format(myname))
    else:
        self.poutput('{!r} has preprocessed SAXS curves'.format(myname))
    pass
    if collbat:
        self.poutput('bat: Attenuated beam evolution is now in OP1 for smoothing and injection')
#END merger()

merger()
