#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables

'''setmerge.py: Merges exactly two sets of .dat-files into a single super-set.
             The result is a new set of .dat-files. Application: To merge a
             "heating branch" with a "cooling branch". An intermediate dwell
             time can be filled with interpolated patterns. More than 2 sets
             can be merged by successive application.
SYNTAX: @@setmerge,<newname>,<branch1>,<branch2>,<dwellframes>,<condense>
EXAMPLE: @@setmerge,mbsy02,bsy02_up*,bsy02_dn*,5,12
    <newname>: filename prefix for the combined set
               If the first character of <newname> is "!", then existing
               <newset> files will be deleted before the new set is
               (re-)combined.
    <branch1>: the group of files which forms the first branch
    <branch2>: the group of files which forms the second branch
    <dwellframes>: optional parameter which specifies the number of
                   artificial frames to be generated between branch1 and
                   branch2.
    <condense>: optional parameter which specifies how many successive input
                points from each curve shall be condensed into a single one.
                (This is useful with small-angle data, if each curve contains
                several thousand data points). "Condensation" calls the
                sascur-command "condense"

Hint: Normalization must be postponed until all data are put together.'''

## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def datfilesfound( test ):
    # True, if there are files matching the pattern
    missingext = len(test.split('.')) == 1
    if missingext:
        # No dot which indicates a file extension
        # Append default extension for data files
        testx = test+'.dat'
    else:
        testx = test
    listoffiles = glob.glob( testx )
    return len(listoffiles) > 0
#END datfilesfound()

def keyav( fracearly, fraclate, key, dest, early, late ):
    # puts the average from early and late weighted by
    # fracearly and fraclate into the dictionary dest
    if key in early:
        if key in late:
            dest[key] = fracearly*early[key] + fraclate*late[key]
            return
    dest[key] = 'n/a' # Indicate failure to average

def mergebranches():
    '''mergebranches does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    # Argument parsing of the first argument is postponed, because it
    # may involve the deletion of old files which shall only be done
    # just before the work starts.

    ### Argument parsing
    argus = len(sys.argv) - 1
    if argus < 3:
        return "Script needs at least 3 arguments"

    ## (2) The front branch
    branch1 = sys.argv[2]
    # If the user has given no extension, then append '.dat'
    test = branch1.split('.')
    if len(test) < 2:
        branch1 = branch1+'.dat'
        pass
    if not datfilesfound( branch1 ):
        return 'branch1: There are no files matching "%s"'%branch1

    ## (3) The tail branch
    branch2 = sys.argv[3]
    # If the user has given no extension, then append '.dat'
    test = branch2.split('.')
    if len(test) < 2:
        branch2 = branch2+'.dat'
        pass
    if not datfilesfound( branch2 ):
        return 'branch2: There are no files matching "%s"'%branch2

    ## (4) The dwellframes
    dwelf = 0 # No extra dwell frames between the branches
    if argus > 3:
        dwelfs = sys.argv[4]
        try:
            dwelf = int(dwelfs)
        except:
            return 'Error converting <dwellframes> "%s" to integer'%dwelfs
        if dwelf < 1:
            return '<dwellframes> must be positive'
        if dwelf > 200:
            return 'That many <dwellframes>???'

    ## (5) Optional curve condensation
    condcur = 1 # no condensation
    if argus > 4:
        conds = sys.argv[5]
        try:
            condcur = int(conds)
        except:
            return 'Error converting <condense> "%s" to integer'%conds
        if condcur < 1:
            return '<condense> must be positive'

    ## (1) the new set
    newname = sys.argv[1]
    if newname == '':
        return '<newname> has not been provided'
    force = newname[0] == '!' # force overwriting
    if force:
        newname = newname[1:]
        if len(newname) < 1:
            return '<newname> has not been provided'
        pass

    checkit = newname+'*.dat'
    newfilesfound = datfilesfound( checkit )
    if force:
        if newfilesfound:
            listoffiles = glob.glob( checkit )
            for i, thefile in enumerate( listoffiles ):
                os.remove( thefile )
            self.poutput('On request old files "%s" have been deleted.'%checkit)
    else:
        if newfilesfound:
            return 'Files "%s" exist, but deleting has not been requested.'%checkit
    ### End argument parsing

    ### start the work
    outindex = 0 # The actual index into the output set
    # sorted list of the first branch
    branchfiles = sorted(glob.glob(branch1))
    for i, thefile in enumerate(branchfiles):
        err = self.ops[0].readascii( thefile )
        if err != '':
            return 'Error reading file "%s" in branch1'%thefile
        if condcur > 1:
            # condense the curve to less points
            cmd('condense,0,%s'%condcur)
        # Make the output file name
        outis = str(outindex)
        while len(outis) < 4:
            # Fill with leading zeroes
            outis = '0'+outis
        outfile = newname+'_'+outis
        # generate the next file in the merged set
        self.poutput('Writing %s'%outfile)
        err = self.ops[0].writeascii( outfile, True ) # overwrite
        if err != '':
            return err
        outindex = outindex + 1
    # branch1 has finished.

    # Continuity management
    attrib = self.ops[0].env
    uptime = attrib['elapsed']
    cycle = attrib['exposure'] + attrib['latency']
    upframe = attrib['frame']

    # branch2:
    branchfiles = sorted(glob.glob(branch2))
    if dwelf > 0:
        # Insert frames bridging the dwell time
        thefile = branchfiles[0] # first frame in the 2nd set
        err = self.ops[1].readascii( thefile )
        if err != '':
            return 'Error reading file "%s" in branch1'%thefile
        if condcur > 1:
            # condense the curve to less points
            cmd('condense,1,%s'%condcur)

        # Check compatibility
        if self.ops[0].n != self.ops[1].n:
            return 'Curves in branch1 and branch2 have different lengths'
        ghost = copy.deepcopy(self.ops[1])
        span = ghost.n
        sufra = dwelf+1
        be = self.ops[0]
        bl = self.ops[1]
        for i in range(1,sufra):
            ghost.env['elapsed'] = uptime  + i*cycle
            ghost.env['frame']   = upframe + i
            # Sliding weight of average with fe and fl being the fractions
            # for weighting
            fe = (sufra -i)/float(sufra)
            fl = i/float(sufra)
            keyav( fe, fl, 'bats', ghost.env, be.env, bl.env )
            # ghost.env['bats'] = early*be.env['bats'] + late*bl.env['bats']
            keyav( fe, fl, 'bat',  ghost.env, be.env, bl.env )
            keyav( fe, fl, 'emut', ghost.env, be.env, bl.env )
            keyav( fe, fl, 'mut',  ghost.env, be.env, bl.env )
            keyav( fe, fl, 'T',    ghost.env, be.env, bl.env )
            ghost.y[:span] = fe*be.y[:span] + fl*bl.y[:span]
            # x-values remain the same

            # Make the output file name
            outis = str(outindex)
            while len(outis) < 4:
                # Fill with leading zeroes
                outis = '0'+outis
            outfile = newname+'_'+outis
            # generate the next file in the merged set
            self.poutput('Writing %s'%outfile)
            err = ghost.writeascii( outfile, True )
            if err != '':
                return err
            outindex = outindex + 1
            pass
        # Continuity management
        attrib  = ghost.env
        uptime  = attrib['elapsed']
        upframe = attrib['frame']
        pass # artificial frames bridging the dwell time

    ## Go on after branch1 or the generation of dwell-time bridge frames
    for i, thefile in enumerate(branchfiles):
        err = self.ops[1].readascii( thefile )
        if err != '':
            return 'Error reading file "%s" in branch2'%thefile
        if condcur > 1:
            # condense the curve to less points
            cmd('condense,1,%s'%condcur)
        # Make the output file name
        outis = str(outindex)
        while len(outis) < 4:
            # Fill with leading zeroes
            outis = '0'+outis
        outfile = newname+'_'+outis
        # Continuity of respective attributes -- i starts from 0
        self.ops[1].env['elapsed'] = uptime  + (i+1)*cycle
        self.ops[1].env['frame']   = upframe + i + 1

        # generate the next file in the merged set
        self.poutput('Writing %s'%outfile)
        err = self.ops[1].writeascii( outfile, True )
        if err != '':
            return err
        outindex = outindex + 1
    # branch2 has finished.

    # Erase used operands
    self.ops[0].clean()
    self.ops[1].clean()
    self.poutput('New frameset "%s" by merging of branches "%s" and "%s"'%(newname,branch1,branch2))
    return ''
#END mergebranches()

# END script procedures

# Main:
err = mergebranches()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()



